# Object hierarchy:
# Artifact (set, get methods)
#   MemArtifact
#   FilesystemArtifact (adds open_file, new_file, etc method)
#     LocalArtifact
#     WandbArtifact

import typing

from . import weave_types as types
from . import ref_base


class Artifact:
    def set(self, key: str, type_: types.Type, obj: typing.Any) -> "ArtifactRef":
        raise NotImplementedError

    def get(self, key: str, type_: types.Type) -> typing.Any:
        raise NotImplementedError


class ArtifactRef(ref_base.Ref):
    def __init__(
        self,
        artifact: Artifact,
        path: typing.Optional[str],
        type: typing.Optional[types.Type] = None,
        obj: typing.Optional[typing.Any] = None,
        extra: typing.Optional[list[str]] = None,
        # Denotes that the ref should be saved as a path ref (just the path and extra)
        # without the full uri, because its a reference within an artifact
        serialize_as_path_ref: bool = False,
    ):
        self.artifact = artifact
        self.path = path
        self.serialize_as_path_ref = serialize_as_path_ref
        super().__init__(obj=obj, type=type, extra=extra)

    def without_extra(self, new_type: typing.Optional[types.Type]) -> "ArtifactRef":
        return self.__class__(
            artifact=self.artifact, path=self.path, type=new_type, obj=None, extra=None
        )

    def with_extra(
        self, new_type: typing.Optional[types.Type], obj: typing.Any, extra: list[str]
    ) -> "ArtifactRef":
        new_extra = self.extra
        if self.extra is None:
            new_extra = []
        else:
            new_extra = self.extra.copy()
        new_extra += extra
        return self.__class__(
            artifact=self.artifact,
            path=self.path,
            type=new_type,
            obj=obj,
            extra=new_extra,
        )

    def local_ref_str(self) -> str:
        raise NotImplementedError



import typing
import pyarrow.compute as pc
import pyarrow as pa

from ..decorator_arrow_op import arrow_op
from .. import weave_types as types
from ..api import op
from ..arrow.list_ import ArrowWeaveList, ArrowWeaveListType
from .. import timestamp as weave_timestamp
from . import util

ARROW_WEAVE_LIST_TIMESTAMP_TYPE = ArrowWeaveListType(types.Timestamp())
ARROW_WEAVE_LIST_BOOLEAN_TYPE = ArrowWeaveListType(types.Boolean())
ARROW_WEAVE_LIST_TIMEDELTA_TYPE = ArrowWeaveListType(types.TimeDelta())

binary_input_type = {
    "self": ARROW_WEAVE_LIST_TIMESTAMP_TYPE,
    "other": types.UnionType(types.Timestamp(), ARROW_WEAVE_LIST_TIMESTAMP_TYPE),
}


@arrow_op(
    name="ArrowWeaveListDate-toNumber",
    input_type={"self": ArrowWeaveListType(types.optional(types.Timestamp()))},
    output_type=ArrowWeaveListType(types.optional(types.Int())),
)
def to_number(self):
    return ArrowWeaveList(
        pc.milliseconds_between(pa.scalar(0, self._arrow_data.type), self._arrow_data),
        types.optional(types.Int()),
        self._artifact,
    )


def adjust_multiple_s(multiple_s: float) -> typing.Tuple[float, str]:
    unit = "second"

    # Critical: Arrow silently fails if the unit is < 1. In such cases,
    # we need to convert to make a conversion
    if multiple_s < 1:
        unit = "nanosecond"
        multiple_s = int(multiple_s * 1e9)

    return multiple_s, unit


@arrow_op(
    name="ArrowWeaveListDate-floor",
    input_type={"self": ArrowWeaveListType(types.optional(types.Timestamp()))},
    output_type=ArrowWeaveListType(types.optional(types.Timestamp())),
)
def floor(self, multiple_s: float):
    multiple_s, unit = adjust_multiple_s(multiple_s)
    return ArrowWeaveList(
        pc.floor_temporal(self._arrow_data, multiple=multiple_s, unit=unit),
        types.optional(types.Timestamp()),
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListDate-ceil",
    input_type={"self": ArrowWeaveListType(types.optional(types.Timestamp()))},
    output_type=ArrowWeaveListType(types.optional(types.Timestamp())),
)
def ceil(self, multiple_s: float):
    multiple_s, unit = adjust_multiple_s(multiple_s)
    return ArrowWeaveList(
        pc.ceil_temporal(
            self._arrow_data,
            multiple=multiple_s,
            unit=unit,
            ceil_is_strictly_greater=True,
        ),
        types.optional(types.Timestamp()),
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListDate-__lt__",
    input_type=binary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def le(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.less(self._arrow_data, other), types.Boolean(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListDate-__le__",
    input_type=binary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def lt(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.less_equal(self._arrow_data, other), types.Boolean(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListDate-__gt__",
    input_type=binary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def gt(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.greater(self._arrow_data, other), types.Boolean(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListDate-__ge__",
    input_type=binary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def ge(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.greater_equal(self._arrow_data, other), types.Boolean(), self._artifact
    )


@op(
    name="ArrowWeaveListDate-min",
    input_type={"self": ArrowWeaveListType(types.optional(types.Timestamp()))},
    output_type=types.optional(types.Timestamp()),
)
def timestamp_min(self):
    array = self._arrow_data_asarray_no_tags()
    return pc.min(array).as_py()


@op(
    name="ArrowWeaveListDate-max",
    input_type={"self": ArrowWeaveListType(types.optional(types.Timestamp()))},
    output_type=types.optional(types.Timestamp()),
)
def timestamp_max(self):
    array = self._arrow_data_asarray_no_tags()
    return pc.max(array).as_py()


@arrow_op(
    name="ArrowWeaveListDate-sub",
    input_type={
        "self": ArrowWeaveListType(types.optional(types.Timestamp())),
        "other": types.UnionType(types.Timestamp(), ARROW_WEAVE_LIST_TIMESTAMP_TYPE),
    },
    output_type=ArrowWeaveListType(types.TimeDelta()),
)
def sub(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.subtract(self._arrow_data, other), types.TimeDelta(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListDate-add",
    input_type={
        "self": ArrowWeaveListType(types.optional(types.Timestamp())),
        "other": types.UnionType(types.TimeDelta(), ARROW_WEAVE_LIST_TIMEDELTA_TYPE),
    },
    output_type=ArrowWeaveListType(types.optional(types.Timestamp())),
)
def add(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.add(self._arrow_data, other),
        types.optional(types.Timestamp()),
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListTimeDelta-totalSeconds",
    input_type={"td": ArrowWeaveListType(types.TimeDelta())},
    output_type=ArrowWeaveListType(types.Number()),
)
def timedelta_total_seconds(td):
    new_arrow_data = pc.divide(pc.cast(td._arrow_data, pa.int64()), 1e6)
    return ArrowWeaveList(new_arrow_data, types.Number(), td._artifact)



import typing
from typing import Mapping
import inspect

from . import graph_client_context
from . import context_state
from . import run_context
from . import object_context
from . import storage
from . import ref_base
from . import weave_types as types
from . import box

if typing.TYPE_CHECKING:
    from .op_def import OpDef


def _deref_all(obj: typing.Any):
    if isinstance(obj, dict):
        return {k: _deref_all(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_deref_all(v) for v in obj]
    elif isinstance(obj, ref_base.Ref):
        return obj.get()
    return obj


def _auto_publish(obj: typing.Any, output_refs: typing.List[ref_base.Ref]):
    import numpy as np

    ref = storage.get_ref(obj)
    if ref:
        output_refs.append(ref)
        return ref

    if isinstance(obj, dict):
        return {k: _auto_publish(v, output_refs) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_auto_publish(v, output_refs) for v in obj]
    weave_type = types.TypeRegistry.type_of(obj)
    if weave_type == types.UnknownType():
        return f"<UnknownType: {type(obj)}>"
    if not (
        types.is_custom_type(weave_type) or isinstance(weave_type, types.ObjectType)
    ):
        return obj

    client = graph_client_context.require_graph_client()
    if not ref:
        ref = client.save_object(obj, f"{obj.__class__.__name__}", "latest")

    output_refs.append(ref)
    return ref


def auto_publish(obj: typing.Any) -> typing.Tuple[typing.Any, list]:
    refs: typing.List[ref_base.Ref] = []
    return _auto_publish(obj, refs), refs


def execute_op(op_def: "OpDef", inputs: Mapping[str, typing.Any]):
    mon_span_inputs = {**inputs}
    client = graph_client_context.get_graph_client()
    if client is not None and context_state.eager_mode() and op_def.location:
        op_def_ref = storage._get_ref(op_def)
        if not client.ref_is_own(op_def_ref):
            op_def_ref = client.save_object(op_def, f"{op_def.name}", "latest")
        mon_span_inputs, refs = auto_publish(inputs)

        # Memoization disabled for now.
        # found_run = client.find_op_run(str(op_def_ref), mon_span_inputs)
        # if found_run:
        #     return found_run.output

        parent_run = run_context.get_current_run()
        # if not parent_run:
        #     print("Running ", op_def.name)
        run = client.create_run(str(op_def_ref), parent_run, mon_span_inputs, refs)
        try:
            with run_context.current_run(run):
                res = op_def.resolve_fn(**inputs)
        except BaseException as e:
            print("Error running ", op_def.name)
            client.fail_run(run, e)
            raise
        if isinstance(res, box.BoxedNone):
            res = None

        # Don't use asyncio.iscoroutine. It returns True for non-async
        # generators sometimes. Use inspect.iscoroutine instead.
        if inspect.iscoroutine(res):

            async def _run_async():
                try:
                    awaited_res = res
                    with object_context.object_context():
                        with run_context.current_run(run):
                            # Need to do this in a loop for some reason to handle
                            # async streaming openai. Like we get two co-routines
                            # in a row.
                            while inspect.iscoroutine(awaited_res):
                                awaited_res = await awaited_res
                    output, output_refs = auto_publish(awaited_res)
                    # TODO: boxing enables full ref-tracking of run outputs
                    # to other run inputs, but its not working yet.
                    # output = box.box(output)
                    client.finish_run(run, output, output_refs)
                    if not parent_run:
                        print("🍩 View call:", run.ui_url)
                    return _deref_all(output)
                except Exception as e:
                    client.fail_run(run, e)
                    raise

            return _run_async()
        else:
            output, output_refs = auto_publish(res)
            # TODO: boxing enables full ref-tracking of run outputs
            # to other run inputs, but its not working yet.
            # output = box.box(output)

            client.finish_run(run, output, output_refs)
            if not parent_run:
                print("🍩 View call:", run.ui_url)
            res = _deref_all(output)

    else:
        res = op_def.resolve_fn(**inputs)

    return res



import typing
import datetime
import inspect
from fastapi import FastAPI, Header, HTTPException, Depends
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from typing import Optional

try:
    from typing import Annotated
# Support python 3.8
except ImportError:
    from typing_extensions import Annotated  # type: ignore

from . import weave_types as types
from . import op_args
from . import weave_pydantic
from . import op_def
from . import cache
from .monitoring import monitor
from .wandb_api import WandbApiAsync, wandb_api_context, WandbApiContext

from .artifact_wandb import WandbArtifactRef

key_cache: cache.LruTimeWindowCache[
    str, typing.Optional[bool]
] = cache.LruTimeWindowCache(datetime.timedelta(minutes=5))

api: Optional[WandbApiAsync] = None


def wandb_auth(
    entity: str,
) -> typing.Callable[
    [typing.Optional[str]], typing.Coroutine[typing.Any, typing.Any, bool]
]:
    async def auth_inner(key: Annotated[Optional[str], Depends(api_key)]) -> bool:
        global api
        if api is None:
            api = WandbApiAsync()
        if key is None:
            raise HTTPException(status_code=401, detail="Missing API Key")
        if len(key.split("-")[-1]) != 40:
            raise HTTPException(status_code=401, detail="Invalid API Key")

        authed = key_cache.get(key)
        if isinstance(authed, bool):
            return authed
        authed = await api.can_access_entity(entity, api_key=key)
        if not authed:
            raise HTTPException(status_code=403, detail="Permission Denied")
        key_cache.set(key, authed)
        return authed

    return auth_inner


def api_key(
    credentials: Annotated[
        Optional[HTTPBasicCredentials],
        Depends(
            HTTPBasic(
                auto_error=False,
                description="Set your username to api and password to a W&B API Key",
            )
        ),
    ],
    x_wandb_api_key: Annotated[
        Optional[str], Header(description="Optional W&B API Key")
    ] = None,
) -> Optional[str]:
    if x_wandb_api_key:
        return x_wandb_api_key
    elif credentials and credentials.password:
        return credentials.password
    else:
        return None


def object_method_app(
    obj_ref: WandbArtifactRef,
    method_name: typing.Optional[str] = None,
    auth_entity: typing.Optional[str] = None,
) -> FastAPI:
    # Import weaveflow to trigger eager mode and ensure we have weaveflow weave
    # types loaded.
    from weave import weaveflow

    obj = obj_ref.get()
    obj_weave_type = types.TypeRegistry.type_of(obj)
    if not isinstance(obj_weave_type, types.ObjectType):
        raise ValueError(
            f"Expected an object (created by @weave.type()), got {obj_weave_type}"
        )

    op_attrs: dict[str, op_def.OpDef] = {
        attr: value
        for attr, value in obj.__class__.__dict__.items()
        if isinstance(value, op_def.OpDef)
    }
    if not op_attrs:
        raise ValueError("No ops found on object")

    if method_name is None:
        if len(op_attrs) > 1:
            raise ValueError(
                "Multiple ops found on object (%s), must specify method_name argument"
                % ", ".join(op_attrs)
            )
        method_name = next(iter(op_attrs))

    bound_method_op = typing.cast(op_def.OpDef, getattr(obj, method_name))
    if not isinstance(bound_method_op, op_def.OpDef):
        raise ValueError(f"Expected an op, got {bound_method_op}")

    bound_method_op_args = bound_method_op.input_type
    if not isinstance(bound_method_op_args, op_args.OpNamedArgs):
        raise ValueError("predict op must have named args")

    arg_types = bound_method_op_args.weave_type().property_types
    del arg_types["self"]

    Item = weave_pydantic.weave_type_to_pydantic(arg_types, name="Item")

    dependencies = []
    if auth_entity:
        dependencies.append(Depends(wandb_auth(auth_entity)))

    app = FastAPI(dependencies=dependencies)

    @app.post(f"/{method_name}", summary=method_name)
    async def method_route(item: Item) -> dict:  # type: ignore
        if inspect.iscoroutinefunction(bound_method_op.raw_resolve_fn):
            result = await bound_method_op(**item.dict())  # type: ignore
        else:
            result = bound_method_op(**item.dict())  # type: ignore
        return {"result": result}

    return app



import json
import os
import re
import subprocess
import sys
import typing


def execute(
    args: list[str],
    timeout: typing.Optional[float] = None,
    cwd: typing.Optional[str] = None,
    env: typing.Optional[dict[str, str]] = None,
    input: typing.Optional[str] = None,
    capture: bool = True,
) -> typing.Any:
    process = subprocess.Popen(
        args,
        stdout=subprocess.PIPE if capture else sys.stdout,
        stderr=subprocess.PIPE if capture else sys.stderr,
        stdin=subprocess.PIPE if capture else sys.stdin,
        universal_newlines=True,
        env=env or os.environ.copy(),
        cwd=cwd,
    )
    out, err = process.communicate(timeout=timeout, input=input)
    if process.returncode != 0:
        raise ValueError(f"Command failed: {err or ''}")

    if not capture:
        return None

    try:
        return json.loads(out)
    except json.JSONDecodeError:
        raise ValueError(f"Failed to parse JSON from command: {out}")


def safe_name(name: str) -> str:
    """The name must use only lowercase alphanumeric characters and dashes,
    cannot begin or end with a dash, and cannot be longer than 63 characters."""
    fixed_name = re.sub(r"[^a-z0-9-]", "-", name.lower()).strip("-")
    if len(fixed_name) == 0:
        return "weave-op"
    elif len(fixed_name) > 63:
        return fixed_name[:63]
    else:
        return fixed_name



from .. import weave_types as types
from ..language_features.tagging import make_tag_getter_op

group_tag_getter_op = make_tag_getter_op.make_tag_getter_op(
    "groupKey", types.Any(), op_name="group-groupkey"
)

index_checkpoint_tag_getter_op = make_tag_getter_op.make_tag_getter_op(
    "indexCheckpoint", types.Int(), op_name="tag-indexCheckpoint"
)

join_obj_getter_op = make_tag_getter_op.make_tag_getter_op(
    "joinObj", types.Any(), types.Any(), "tag-joinObj"
)



import re
import contextlib
import dataclasses
import os
import functools
import tempfile
import typing
import requests
import logging

from wandb import Artifact
from wandb.apis.public import api as wb_public
from wandb.sdk.lib.hashutil import hex_to_b64_id, b64_to_hex_id

from . import urls
from . import uris
from . import util
from . import errors
from . import wandb_client_api
from . import file_base
from . import file_util

from . import weave_types as types
from . import artifact_fs
from . import filesystem
from . import memo
from . import eager
from . import graph_client_context
from .wandb_interface import wandb_artifact_pusher
from . import engine_trace

from urllib import parse

if typing.TYPE_CHECKING:
    from weave.wandb_interface.wandb_lite_run import InMemoryLazyLiteRun
    from .run_streamtable_span import RunStreamTableSpan


quote_slashes = functools.partial(parse.quote, safe="")

DEFAULT_WEAVE_OBJ_PROJECT = "weave"


class WandbArtifactManifestEntry(typing.TypedDict):
    digest: str
    birthArtifactID: typing.Optional[str]
    ref: typing.Optional[str]
    size: typing.Optional[int]
    extra: typing.Optional[dict[str, typing.Any]]
    # wandb.sdk.wandb_artifacts.ArtifactManifestEntry also has "local_path"
    # but that doesn't exist in saved manifests


class WandbArtifactManifestV1(typing.TypedDict):
    version: int
    storagePolicy: str
    storagePolicyConfig: typing.Optional[dict[str, typing.Any]]
    contents: typing.Dict[str, WandbArtifactManifestEntry]


# We used to use wandb.sdk.wandb_artifacts.ArtifactManifest directly, but it
# is expensive to construct because it makes new objects for every entry and
# causes a lot of memory/gc churn. This implementation leaves the manifest
# as the raw json dict instead.
@dataclasses.dataclass
class WandbArtifactManifest:
    class StorageLayout:
        V1 = "V1"
        V2 = "V2"

    _manifest_json: WandbArtifactManifestV1

    @property
    def _storage_policy_config(self):
        return self._manifest_json.get("storagePolicyConfig", {})

    @property
    def storage_layout(self):
        return self._storage_policy_config.get(
            "storageLayout", WandbArtifactManifest.StorageLayout.V1
        )

    def get_entry_by_path(
        self, path: str
    ) -> typing.Optional[WandbArtifactManifestEntry]:
        return self._manifest_json["contents"].get(path)

    def get_paths_in_directory(self, path: str) -> typing.List[str]:
        if path == "":
            return list(self._manifest_json["contents"].keys())
        dir_path = path + "/"
        return [
            k for k in self._manifest_json["contents"].keys() if k.startswith(dir_path)
        ]


# TODO: Get rid of this, we have the new wandb api service! But this
# is still used in a couple places.
@memo.memo  # Per-request memo reduces duplicate calls to the API
def get_wandb_read_artifact(path: str):
    tracer = engine_trace.tracer()
    with tracer.trace("get_wandb_read_artifact"):
        try:
            return wandb_client_api.wandb_public_api().artifact(path)
        except wandb_client_api.WandbCommError:
            raise errors.WeaveArtifactVersionNotFound(
                f"Could not find artifact with path {path} in W&B"
            )


def is_valid_version_index(version_index: str) -> bool:
    return bool(re.match(r"^v(0|[1-9][0-9]*)$", version_index))


def get_wandb_read_artifact_uri(path: str):
    tracer = engine_trace.tracer()
    with tracer.trace("get_wandb_read_artifact_uri"):
        art = get_wandb_read_artifact(path)
        return WeaveWBArtifactURI(
            art.name.split(":", 1)[0],
            art.commit_hash,
            art.entity,
            art.project,
        )


def wandb_artifact_dir():
    d = os.path.join(filesystem.get_filesystem_dir(), "wandb_artifacts")
    os.makedirs(d, exist_ok=True)
    return d


@dataclasses.dataclass
class ReadClientArtifactURIResult:
    weave_art_uri: "WeaveWBArtifactURI"
    artifact_type_name: str
    is_deleted: bool


def _art_id_is_client_version_id_mapping(art_id: str) -> bool:
    return len(art_id) == 128 and ":" not in art_id


def _art_id_is_client_collection_and_alias_id_mapping(art_id: str) -> bool:
    return ":" in art_id


def _convert_client_id_to_server_id(art_id: str) -> str:
    query = wb_public.gql(
        """
        query ClientIDMapping($clientID: ID!) {
            clientIDMapping(clientID: $clientID) {
                serverID
            }
        }
    """
    )
    tracer = engine_trace.tracer()
    with tracer.trace("_convert_client_id_to_server_id.execute"):
        res = wandb_client_api.wandb_public_api().client.execute(
            query,
            variable_values={
                "clientID": art_id,
            },
        )
    return b64_to_hex_id(res["clientIDMapping"]["serverID"])


def _collection_and_alias_id_mapping_to_uri(
    client_collection_id: str, alias_name: str
) -> ReadClientArtifactURIResult:
    is_deleted = False
    query = """
    query ArtifactVersionFromIdAlias(
        $id: ID!,
        $aliasName: String!
    ) {
        artifactCollection(id: $id) {
            id
            name
            state
            project {
                id
                name
                entity {
                    id
                    name
                }
            }
            artifactMembership(aliasName: $aliasName) {
                id
                versionIndex
                commitHash
            }
            artifactMemberships(first: 1) {
                edges {
                    node {
                        id
                        versionIndex
                        commitHash
                    }
                }
            }
            defaultArtifactType {
                id
                name
            }
        }
    }
    """

    try:
        tracer = engine_trace.tracer()
        with tracer.trace("_collection_and_alias_id_mapping_to_uri.execute"):
            res = wandb_client_api.query_with_retry(
                query,
                variables={
                    "id": client_collection_id,
                    "aliasName": alias_name,
                },
                num_timeout_retries=1,
            )
    except requests.exceptions.HTTPError as e:
        if e.response is not None and e.response.status_code == 400:
            # This is a special case: the client id corresponds to an artifact that was
            # never uploaded, so the client id doesn't exist in the W&B server.

            collection = None
            logging.warn(
                f"Artifact collection with client id {client_collection_id} not present in W&B server."
            )

        else:
            raise e
    else:
        collection = res["artifactCollection"]

    if collection is None:
        # Note: deleted collections are still returned by the API (with state=DELETED)
        # So a missing collection is a real error (unless it was never uploaded)
        raise errors.WeaveArtifactCollectionNotFound(
            f"Could not find artifact collection with client id {client_collection_id}"
        )
    elif collection["state"] != "READY":  # state is either "DELETED" or "READY"
        # This is a deleted artifact, but we still have a record of it.
        is_deleted = True

    artifact_type_name = res["artifactCollection"]["defaultArtifactType"]["name"]
    artifact_membership = res["artifactCollection"]["artifactMembership"]
    if artifact_membership is None:
        # Here we account for a special case: when the alias is `latest`. It is
        # exceedingly common and expected to reference the latest alias of a
        # collection. This is particularly common when logging tables (but used
        # in many other places as well). Occasionally the user deletes the
        # `latest` alias which is a super easy thing to do in the UI. However,
        # this results in the data not being found. We solved this similarly in
        # the typescript implementation by getting the most recent version in
        # such cases. At one point we had an extended discussion about making
        # `latest` a "calculated" alias, which would solve this issue on
        # gorilla's side, but were unable to reach an actionable conclusion.
        # This is a simple workaround that solves for these cases.
        if (
            alias_name == "latest"
            and len(res["artifactCollection"]["artifactMemberships"]["edges"]) > 0
        ):
            commit_hash = res["artifactCollection"]["artifactMemberships"]["edges"][0][
                "node"
            ]["commitHash"]
        else:
            is_deleted = True
            # If the membership is deleted, then we will not be able to get the commit hash.
            # Here, we can simply use the alias name as the commit hash. If the collection
            # is ever un-deleted or the alias is re-assigned, then the next call will result
            # in a "true" commit hash, ensuring we don't hit a stale cache.
            commit_hash = alias_name
    else:
        commit_hash = artifact_membership["commitHash"]
    entity_name = collection["project"]["entity"]["name"]
    project_name = collection["project"]["name"]
    artifact_name = collection["name"]

    weave_art_uri = WeaveWBArtifactURI(
        artifact_name,
        commit_hash,
        entity_name,
        project_name,
    )

    return ReadClientArtifactURIResult(weave_art_uri, artifact_type_name, is_deleted)


def _version_server_id_to_uri(server_id: str) -> ReadClientArtifactURIResult:
    is_deleted = False
    query = wb_public.gql(
        """
    query ArtifactVersionFromServerId(
        $id: ID!,
    ) {
        artifact(id: $id) {
            id
            state
            commitHash
            versionIndex
            artifactType {
                id
                name
            }
            artifactSequence {
                id
                name
                state
                project {
                    id
                    name
                    entity {
                        id
                        name
                    }
                }
            }
        }
    }
    """
    )
    tracer = engine_trace.tracer()
    with tracer.trace("_version_server_id_to_uri.execute"):
        res = wandb_client_api.wandb_public_api().client.execute(
            query,
            variable_values={"id": hex_to_b64_id(server_id)},
        )

    artifact = res["artifact"]
    if artifact is None:
        # Note: deleted versions are still returned by the API (with state=DELETED)
        # So a missing artifact is a real error.
        raise errors.WeaveArtifactVersionNotFound(
            f"Could not find artifact version with server id {server_id}"
        )
    elif (
        artifact["state"] != "COMMITTED"
    ):  # state is either "DELETED" or "PENDING" or "COMMITTED"
        # This is a deleted artifact, but we still have a record of it
        is_deleted = True

    collection = artifact["artifactSequence"]

    if collection is None:
        # Note: deleted collections are still returned by the API (with state=DELETED)
        # So a missing collection is a real error.
        raise errors.WeaveArtifactCollectionNotFound(
            f"Could not find artifact sequence for artifact version with server id {server_id}"
        )
    elif collection["state"] != "READY":  # state is either "DELETED" or "READY"
        # This is a deleted artifact, but we still have a record of it
        is_deleted = True

    artifact_type_name = artifact["artifactType"]["name"]
    commit_hash = artifact["commitHash"]

    entity_name = collection["project"]["entity"]["name"]
    project_name = collection["project"]["name"]
    artifact_name = collection["name"]

    weave_art_uri = WeaveWBArtifactURI(
        artifact_name,
        commit_hash,
        entity_name,
        project_name,
    )

    return ReadClientArtifactURIResult(weave_art_uri, artifact_type_name, is_deleted)


def get_wandb_read_client_artifact_uri(art_id: str) -> ReadClientArtifactURIResult:
    """art_id may be client_id, seq:alias, or server_id"""
    tracer = engine_trace.tracer()
    with tracer.trace("get_wandb_read_client_artifact_uri"):
        if _art_id_is_client_version_id_mapping(art_id):
            server_id = _convert_client_id_to_server_id(art_id)
            return _version_server_id_to_uri(server_id)
        elif _art_id_is_client_collection_and_alias_id_mapping(art_id):
            client_collection_id, alias_name = art_id.split(":")
            return _collection_and_alias_id_mapping_to_uri(
                client_collection_id, alias_name
            )
        else:
            return _version_server_id_to_uri(art_id)


def get_wandb_read_client_artifact(art_id: str) -> typing.Optional["WandbArtifact"]:
    """art_id may be client_id, seq:alias, or server_id"""
    tracer = engine_trace.tracer()
    with tracer.trace("get_wandb_read_client_artifact"):
        res = get_wandb_read_client_artifact_uri(art_id)
        if res.is_deleted:
            return None
        return WandbArtifact(
            res.weave_art_uri.name, res.artifact_type_name, res.weave_art_uri
        )


class WandbArtifactType(artifact_fs.FilesystemArtifactType):
    def save_instance(self, obj, artifact, name) -> "WandbArtifactRef":
        return WandbArtifactRef(obj, None)


class WandbArtifact(artifact_fs.FilesystemArtifact):
    def __init__(
        self,
        name,
        type=None,
        uri: typing.Optional[
            typing.Union["WeaveWBArtifactURI", "WeaveWBLoggedArtifactURI"]
        ] = None,
    ):
        from . import io_service

        self.io_service = io_service.get_sync_client()
        self.name = name

        # original uri passed, if any. this can include aliases such as latest or best, which do not
        # correspond to specific versions or may change versions.
        self._unresolved_read_artifact_or_client_uri = None

        # resolved version of the URI above. this points to the same artifact as the unresolved URI but
        # includes a specific, immutable version like v4 instead of an alias. this is needed for idempotency
        # of cache keys.
        self._resolved_read_artifact_uri: typing.Optional["WeaveWBArtifactURI"] = None
        self._read_artifact = None
        if not uri:
            self._writeable_artifact = Artifact(
                name, "op_def" if type is None else type
            )
        else:
            # load an existing artifact, this should be read only,
            # TODO: we could technically support writable artifacts by creating a new version?
            self._unresolved_read_artifact_or_client_uri = uri
        self._local_path: dict[str, str] = {}
        # self._path_handlers: dict[str, typing.Any] = {}

    @property
    def branch(self) -> typing.Optional[str]:
        if self._unresolved_read_artifact_or_client_uri is not None:
            branch = self._unresolved_read_artifact_or_client_uri.version

        if branch is not None and not likely_commit_hash(branch):
            return branch

        return None

    @property
    def branch_point(self) -> typing.Optional[artifact_fs.BranchPointType]:
        # all branching is local for now. we could support remote branches
        # but in this initial implementation, assume only LocalArtifacts
        # live on branches.
        return None

    @property
    def _read_artifact_uri(self) -> typing.Optional["WeaveWBArtifactURI"]:
        if self._resolved_read_artifact_uri is not None:
            return self._resolved_read_artifact_uri

        if isinstance(
            self._unresolved_read_artifact_or_client_uri,
            (WeaveWBLoggedArtifactURI, WeaveWBArtifactURI),
        ):
            self._resolved_read_artifact_uri = (
                self._unresolved_read_artifact_or_client_uri.resolved_artifact_uri
            )
        return self._resolved_read_artifact_uri

    @property
    def _ref(self) -> "WandbArtifactRef":
        # existing_ref = ref_base.get_ref(obj)
        # if isinstance(existing_ref, artifact_base.ArtifactRef):
        #     return existing_ref
        if not self.is_saved:
            raise errors.WeaveInternalError("cannot get ref of an unsaved artifact")
        return WandbArtifactRef(self, None, None)

    def _set_read_artifact_uri(self, uri):
        self._read_artifact = None
        self._resolved_read_artifact_uri = None
        self._unresolved_read_artifact_or_client_uri = uri

    # TODO: still using wandb lib for this, but we should switch to the new
    # wandb api service
    @property
    def _saved_artifact(self):
        if self._read_artifact is None:
            uri = self._read_artifact_uri
            path = f"{uri.entity_name}/{uri.project_name}/{uri.name}:{uri.version}"
            self._read_artifact = get_wandb_read_artifact(path)
        return self._read_artifact

    def __repr__(self):
        return "<WandbArtifact %s>" % self.name

    def delete(self) -> None:
        self._saved_artifact.delete(delete_aliases=True)

    def add_artifact_reference(self, uri: str) -> None:
        # The URI for a get node might look like: wandb-artifact:///<entity>/<project>/test8:latest/obj
        # but the SDK add_reference call requires something like:
        # wandb-artifact://41727469666163743a353432353830303535/obj.object.json
        assert uri.startswith("wandb-artifact:///")
        uri_parts = WeaveWBArtifactURI.parse(uri)
        uri_path = f"{uri_parts.entity_name}/{uri_parts.project_name}/{uri_parts.name}:{uri_parts.version}"
        ref_artifact = get_wandb_read_artifact(uri_path)

        # A reference needs to point to a specific existing file in the other artifact.
        # For stream tables obj.object.json should exist. For logged tables, the filename
        # can vary, so we just pick the first file we find.
        filename = "obj.object.json"
        try:
            ref_url = ref_artifact.get_path(filename).ref_url()
        except KeyError:
            for ref_file in ref_artifact.files(None, per_page=1):
                ref_url = ref_artifact.get_path(ref_file.name).ref_url()
                break

        # We have a name collision problem. E.g. A board and a stream table will both have
        # a file named obj.object.json. Add a prefix to distinguish them.
        ref_name = f"{ref_artifact.id}__{filename}"
        self._writeable_artifact.add_reference(ref_url, ref_name)

    @property
    def commit_hash(self) -> str:
        if not self.is_saved:
            raise errors.WeaveInternalError(
                "cannot get commit hash of an unsaved artifact"
            )
        resolved_uri = self._read_artifact_uri
        if resolved_uri is None:
            raise errors.WeaveInternalError("cannot resolve commit hash of artifact")
        version = resolved_uri.version
        if version is None:
            raise errors.WeaveInternalError("resolved uri has no version")
        return version

    @property
    def is_saved(self) -> bool:
        return (
            self._unresolved_read_artifact_or_client_uri is not None
            or self._read_artifact is not None
        )

    @property
    def version(self):
        if not self.is_saved:
            raise errors.WeaveInternalError("cannot get version of an unsaved artifact")

        # Note: W&B Artifact's `version` is often the `v0` alias, which is not what
        # we want here (despite the same name for the property). In Weave world, `version`
        # refers to either: a) a commit hash, or b) a branch name. Therefore, in this
        # case, we explicitly want the commit hash, not the differently-defined `version`.
        return self._saved_artifact.commit_hash

    @property
    def created_at(self):
        raise NotImplementedError()

    def get_other_version(self, version):
        raise NotImplementedError()

    def direct_url(self, path: str) -> typing.Optional[str]:
        if self._read_artifact_uri is None:
            raise errors.WeaveInternalError(
                'cannot get direct url for unsaved artifact"'
            )
        uri = self._read_artifact_uri.with_path(path)
        return self.io_service.direct_url(uri)

    def path(self, name: str) -> str:
        if not self.is_saved or not self._read_artifact_uri:
            raise errors.WeaveInternalError("cannot download of an unsaved artifact")

        uri = self._read_artifact_uri.with_path(name)
        fs_path = self.io_service.ensure_file(uri)
        if fs_path is None:
            # Important to raise FileNotFoundError here, FileSystemArtifactRef.type
            # relies on this.
            raise FileNotFoundError("Path not in artifact %s %s" % (self, name))
        return self.io_service.fs.path(fs_path)

    def size(self, path: str) -> int:
        manifest = self._manifest()
        if manifest is not None:
            manifest_entry = manifest.get_entry_by_path(path)
            if manifest_entry:
                size = manifest_entry["size"]
                if size is None:
                    return 0
                return size

        if path in self._saved_artifact.manifest.entries:
            return self._saved_artifact.manifest.entries[path].size
        return super().size(path)

    @property
    def initial_uri_obj(
        self,
    ) -> typing.Union["WeaveWBArtifactURI", "WeaveWBLoggedArtifactURI"]:
        if not self.is_saved or not self._unresolved_read_artifact_or_client_uri:
            raise errors.WeaveInternalError("cannot get uri of an unsaved artifact")
        return self._unresolved_read_artifact_or_client_uri

    @property
    def uri_obj(self) -> "WeaveWBArtifactURI":
        if not self.is_saved or not self._read_artifact_uri:
            raise errors.WeaveInternalError("cannot get uri of an unsaved artifact")
        return self._read_artifact_uri

    @contextlib.contextmanager
    def new_file(self, path, binary=False):
        if not self._writeable_artifact:
            raise errors.WeaveInternalError("cannot add new file to readonly artifact")
        mode = "w"
        if binary:
            mode = "wb"
        with self._writeable_artifact.new_file(path, mode) as f:
            yield f

    @contextlib.contextmanager
    def writeable_file_path(self, path):
        with tempfile.TemporaryDirectory() as tmpdir:
            full_path = os.path.join(tmpdir, path)
            os.makedirs(os.path.dirname(full_path), exist_ok=True)
            yield full_path
            self._writeable_artifact.add_file(full_path, path)

    @contextlib.contextmanager
    def new_dir(self, path):
        if not self._writeable_artifact:
            raise errors.WeaveInternalError("cannot add new file to readonly artifact")
        with tempfile.TemporaryDirectory() as tmpdir:
            yield os.path.abspath(tmpdir)
            self._writeable_artifact.add_dir(tmpdir, path)

    @contextlib.contextmanager
    def open(self, path, binary=False):
        if not self.is_saved:
            raise errors.WeaveInternalError("cannot load data from an unsaved artifact")
        mode = "r"
        if binary:
            mode = "rb"
        p = self.path(path)
        with file_util.safe_open(p, mode) as f:
            yield f

    # def get_path_handler(self, path, handler_constructor):
    #     handler = self._path_handlers.get(path)
    #     if handler is None:
    #         handler = handler_constructor(self, path)
    #         self._path_handlers[path] = handler
    #     return handler

    def read_metadata(self):
        raise NotImplementedError()

    def write_metadata(self, dirname):
        raise NotImplementedError()

    def save(
        self,
        project: str = DEFAULT_WEAVE_OBJ_PROJECT,
        entity_name: typing.Optional[str] = None,
        branch: typing.Optional[str] = None,
        artifact_collection_exists: bool = False,
        *,
        _lite_run: typing.Optional["InMemoryLazyLiteRun"] = None,
    ):
        # for handler in self._path_handlers.values():
        #     handler.close()
        additional_aliases = [] if branch is None else [branch]
        res = wandb_artifact_pusher.write_artifact_to_wandb(
            self._writeable_artifact,
            project,
            entity_name,
            additional_aliases,
            artifact_collection_exists=artifact_collection_exists,
            _lite_run=_lite_run,
        )
        version = res.version_str if branch is None else branch
        self._set_read_artifact_uri(
            WeaveWBArtifactURI(
                name=res.artifact_name,
                version=version,
                entity_name=res.entity_name,
                project_name=res.project_name,
            )
        )

    def _manifest(self) -> typing.Optional[WandbArtifactManifest]:
        if self._read_artifact_uri is None:
            raise errors.WeaveInternalError(
                'cannot get path info for unsaved artifact"'
            )
        return self.io_service.manifest(self._read_artifact_uri)

    def digest(self, path: str) -> typing.Optional[str]:
        manifest_entry = self._manifest_entry(path)
        if manifest_entry is not None:
            return manifest_entry["digest"]
        return None

    def _manifest_entry(self, path: str) -> typing.Optional[WandbArtifactManifestEntry]:
        manifest = self._manifest()
        if manifest is not None:
            manifest_entry = manifest.get_entry_by_path(path)
            if manifest_entry is not None:
                return manifest_entry
        return None

    def _path_info(
        self, path: str
    ) -> typing.Optional[
        typing.Union[
            "artifact_fs.FilesystemArtifactFile",
            "artifact_fs.FilesystemArtifactDir",
            "artifact_fs.FilesystemArtifactRef",
        ]
    ]:
        manifest = self._manifest()
        if manifest is None:
            return None
        manifest_entry = manifest.get_entry_by_path(path)
        if manifest_entry is not None:
            # This is not a WeaveURI! Its the artifact reference style used
            # by the W&B Artifacts/media layer.
            ref_prefix = "wandb-artifact://"
            ref = manifest_entry.get("ref")
            if ref and ref.startswith(ref_prefix):
                # This is a reference to another artifact
                art_id, target_path = ref[len(ref_prefix) :].split("/", 1)
                art = get_wandb_read_client_artifact(art_id)
                # this should be None when the requested artifact is deleted from the server.
                # we want to return None in this case so that the caller can handle it.
                if art is None:
                    return None
                return artifact_fs.FilesystemArtifactFile(art, target_path)
            # This is a file
            return artifact_fs.FilesystemArtifactFile(self, path)

        # This is not a file, assume its a directory. If not, we'll return an empty result.
        dir_paths = manifest.get_paths_in_directory(path)
        sub_dirs: dict[str, file_base.SubDir] = {}
        files = {}
        total_size = 0
        sub_dir_sizes = {}
        cwd = os.getcwd()
        for entry_path in dir_paths:
            entry = manifest.get_entry_by_path(entry_path)
            path_size = (
                entry["size"] if entry is not None and entry["size"] is not None else 0
            )
            total_size += path_size

            # this used to be os.path.relpath but that called os.getcwd() every time
            # that turned out to be a bottleneck in production for artifacts with many
            # dir paths, so we use our own implementation that takes the cwd as input
            # and doesn't need to ever call os.getcwd()
            rel_path = util.relpath_no_syscalls(entry_path, path, cwd)
            rel_path_parts = rel_path.split("/")
            if len(rel_path_parts) == 1:
                files[rel_path_parts[0]] = artifact_fs.FilesystemArtifactFile(
                    self,
                    entry_path,
                )
            else:
                dir_name = rel_path_parts[0]
                if dir_name not in sub_dirs:
                    dir_ = file_base.SubDir(entry_path, 0, {}, {})
                    sub_dir_sizes[dir_name] = 0
                    sub_dirs[dir_name] = dir_
                sub_dir_sizes[dir_name] += path_size
                dir_ = sub_dirs[dir_name]
                if len(rel_path_parts) == 2:
                    dir_files = typing.cast(dict, dir_.files)
                    dir_files[rel_path_parts[1]] = artifact_fs.FilesystemArtifactFile(
                        self,
                        entry_path,
                    )
                else:
                    dir_.dirs[rel_path_parts[1]] = 1
        if not sub_dirs and not files:
            return None
        for dir_name, dir_ in sub_dirs.items():
            dir_.size = sub_dir_sizes[dir_name]
        return artifact_fs.FilesystemArtifactDir(
            self, path, total_size, sub_dirs, files
        )

    def _get_file_paths(self) -> list[str]:
        manifest = self._manifest()
        if manifest is None:
            raise errors.WeaveInternalError("No manifest when fetching file paths")
        return manifest.get_paths_in_directory("")

    @property
    def metadata(self) -> artifact_fs.ArtifactMetadata:
        mutable_metadata = {}
        readonly_metadata = {}
        if (
            hasattr(self, "_writeable_artifact")
            and self._writeable_artifact is not None
        ):
            mutable_metadata = self._writeable_artifact.metadata
        if self.is_saved:
            readonly_metadata = self._saved_artifact.metadata
        return artifact_fs.ArtifactMetadata(mutable_metadata, readonly_metadata)


WandbArtifactType.instance_classes = WandbArtifact


class WandbArtifactRef(artifact_fs.FilesystemArtifactRef):
    artifact: WandbArtifact

    def versions(self) -> list[artifact_fs.FilesystemArtifactRef]:
        # TODO: implement versions on wandb artifact
        return [self]

    @classmethod
    def from_uri(cls, uri: uris.WeaveURI) -> "WandbArtifactRef":
        if not isinstance(uri, (WeaveWBArtifactURI, WeaveWBLoggedArtifactURI)):
            raise errors.WeaveInternalError(
                f"Invalid URI class passed to WandbArtifactRef.from_uri: {type(uri)}"
            )
        # TODO: potentially need to pass full entity/project/name instead
        return cls(
            WandbArtifact(uri.name, uri=uri),
            path=uri.path,
            extra=uri.extra,
        )

    @property
    def ui_url(self):
        root_type = self.type.root_type_class()
        from .op_def_type import OpDefType

        if issubclass(root_type, OpDefType):
            return urls.op_version_path(
                self.artifact.uri_obj.entity_name,
                self.artifact.uri_obj.project_name,
                self.artifact.uri_obj.name,
                self.artifact.uri_obj.version,
            )
        else:
            return urls.object_version_path(
                self.artifact.uri_obj.entity_name,
                self.artifact.uri_obj.project_name,
                self.artifact.uri_obj.name,
                self.artifact.uri_obj.version,
            )

        # Before Tim's Weaveflow changes
        # return f"http://localhost:3000/browse2/{self.artifact.uri_obj.entity_name}/{self.artifact.uri_obj.project_name}/{self.type.root_type_class().name}/{self.artifact.uri_obj.name}/{self.artifact.uri_obj.version}"


types.WandbArtifactRefType.instance_class = WandbArtifactRef
types.WandbArtifactRefType.instance_classes = WandbArtifactRef

WandbArtifact.RefClass = WandbArtifactRef


def is_hex_string(s: str) -> bool:
    try:
        int(s, 16)
        return True
    except ValueError:
        return False


WANDB_COMMIT_HASH_LENGTH = 20


def likely_commit_hash(version: str) -> bool:
    # This is the heuristic we use everywhere to tell if a version is an alias or not
    # if self.version:
    return is_hex_string(version) and len(version) == WANDB_COMMIT_HASH_LENGTH


# Used to refer to objects stored in WB Artifacts. This URI must not change and
# matches the existing artifact schemes
@dataclasses.dataclass
class WeaveWBArtifactURI(uris.WeaveURI):
    SCHEME = "wandb-artifact"
    entity_name: str
    project_name: str
    netloc: typing.Optional[str] = None
    path: typing.Optional[str] = None
    extra: typing.Optional[list[str]] = None

    # resolved version of the artifact uri, in which aliases are replaced with specific versions
    _resolved_artifact_uri: typing.Optional["WeaveWBArtifactURI"] = None

    @classmethod
    def from_parsed_uri(
        cls,
        uri: str,
        schema: str,
        netloc: str,
        path: str,
        params: str,
        query: dict[str, list[str]],
        fragment: str,
    ):
        parts = path.strip("/").split("/")
        parts = [parse.unquote(part) for part in parts]
        if len(parts) < 3:
            raise errors.WeaveInvalidURIError(f"Invalid WB Artifact URI: {uri}")
        entity_name = parts[0]
        project_name = parts[1]

        if ":" in parts[2]:
            name, version = parts[2].split(":", 1)
        else:
            name = parts[2]
            version = None

        file_path: typing.Optional[str] = None
        if len(parts) > 3:
            file_path = "/".join(parts[3:])

        extra: typing.Optional[list[str]] = None
        if fragment:
            extra = fragment.split("/")
        return cls(
            name,
            version,
            entity_name,
            project_name,
            netloc,
            file_path,
            extra,
        )

    @classmethod
    def parse(cls: typing.Type["WeaveWBArtifactURI"], uri: str) -> "WeaveWBArtifactURI":
        return super().parse(uri)  # type: ignore

    def __str__(self) -> str:
        netloc = self.netloc or ""
        uri = (
            f"{self.SCHEME}://"
            f"{quote_slashes(netloc)}/"
            f"{quote_slashes(self.entity_name)}/"
            f"{quote_slashes(self.project_name)}/"
            f"{quote_slashes(self.name)}:{self.version if self.version else ''}"
        )
        if self.path:
            uri += f"/{quote_slashes(self.path)}"
        if self.extra:
            uri += f"#{'/'.join([quote_slashes(e) for e in self.extra])}"
        return uri

    def with_path(self, path: str) -> "WeaveWBArtifactURI":
        return WeaveWBArtifactURI(
            self.name,
            self.version,
            self.entity_name,
            self.project_name,
            self.netloc,
            path,
            self.extra,
        )

    @property
    def resolved_artifact_uri(self) -> "WeaveWBArtifactURI":
        if self.version and likely_commit_hash(self.version):
            return self
        if self._resolved_artifact_uri is None:
            path = f"{self.entity_name}/{self.project_name}/{self.name}"
            if self.version:
                path += f":{self.version}"
            resolved_artifact_uri = get_wandb_read_artifact_uri(path)
            self._resolved_artifact_uri = resolved_artifact_uri
        return self._resolved_artifact_uri

    def to_ref(self) -> WandbArtifactRef:
        return WandbArtifactRef.from_uri(self)


@dataclasses.dataclass
class WeaveWBLoggedArtifactURI(uris.WeaveURI):
    SCHEME = "wandb-logged-artifact"
    # wandb-logged-artifact://afdsjaksdjflkasjdf12341234hv12h3v4k12j3hv41kh4v1423k14v1k2j3hv1k2j3h4v1k23h4v:[version|latest]/path
    #  scheme                                 name                                                              version      path
    path: typing.Optional[str] = None
    extra: typing.Optional[list[str]] = None

    # private attrs

    # resolved version of the artifact uri, in which aliases are replaced with specific versions
    _resolved_artifact_uri: typing.Optional[WeaveWBArtifactURI] = None

    @classmethod
    def from_parsed_uri(
        cls,
        uri: str,
        schema: str,
        netloc: str,
        path: str,
        params: str,
        query: dict[str, list[str]],
        fragment: str,
    ):
        path = parse.unquote(path.strip("/"))
        spl_netloc = netloc.split(":")
        if len(spl_netloc) == 1:
            name = spl_netloc[0]
            version = None
        elif len(spl_netloc) == 2:
            name, version = spl_netloc
        else:
            raise errors.WeaveInvalidURIError(f"Invalid WB Client Artifact URI: {uri}")

        return WeaveWBLoggedArtifactURI(
            name=name,
            version=version,
            path=path,
        )

    def __str__(self) -> str:
        netloc = self.name
        if self.version:
            netloc += f":{self.version}"
        path = self.path or ""
        if path != "":
            path = f"/{path}"
        return f"{self.SCHEME}://{netloc}{path}"

    @property
    def resolved_artifact_uri(self) -> WeaveWBArtifactURI:
        if self._resolved_artifact_uri is None:
            art_id = self.name
            if self.version:
                # version is the alias or version index string
                art_id += f":{self.version}"
            res = get_wandb_read_client_artifact_uri(art_id)
            self._resolved_artifact_uri = res.weave_art_uri
        return self._resolved_artifact_uri.with_path(self.path or "")

    @property
    def entity_name(self) -> str:
        return self.resolved_artifact_uri.entity_name

    @property
    def project_name(self) -> str:
        return self.resolved_artifact_uri.project_name

    def with_path(self, path: str) -> "WeaveWBLoggedArtifactURI":
        return WeaveWBLoggedArtifactURI(
            self.name,
            self.version,
            path,
        )

    @classmethod
    def parse(cls, uri: str) -> "WeaveWBLoggedArtifactURI":
        scheme, netloc, path, params, query_s, fragment = parse.urlparse(uri)
        return cls.from_parsed_uri(
            uri, scheme, netloc, path, params, parse.parse_qs(query_s), fragment
        )

    def to_ref(self) -> WandbArtifactRef:
        return WandbArtifactRef.from_uri(self)


# This is a wrapper around an artifact that acts like a list of files.
# It fetchs a file from the manifest on __getItem__ and can return a count without fetching all files
@dataclasses.dataclass
class FilesystemArtifactFileIterator(list[artifact_fs.FilesystemArtifactFile]):
    data: list[str]
    artifact: WandbArtifact
    idx: int = 0

    def __init__(self, artifact: WandbArtifact, data: list[str] = []):
        self.data = data if len(data) > 0 else artifact._get_file_paths()
        self.artifact = artifact
        self.idx = 0

    def __getitem__(self, key):
        path_or_paths = self.data[key]
        if isinstance(path_or_paths, str):
            return self.artifact._path_info(path_or_paths)
        elif isinstance(path_or_paths, list):
            return FilesystemArtifactFileIterator(self.artifact, path_or_paths)
        raise errors.WeaveInternalError(
            "Invalid key in FilesystemArtifactFileIterator __getItem__"
        )

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self):
        return self

    def __next__(self):
        if self.idx >= len(self.data):
            raise StopIteration
        current_element = self.__getitem__(self.idx)
        self.idx += 1
        return current_element



import pyarrow.compute as pc
import pyarrow as pa

from ..decorator_op import op
from ..decorator_arrow_op import arrow_op
from .. import weave_types as types
from . import util
from ..arrow.list_ import ArrowWeaveList, ArrowWeaveListType

unary_input_type = {
    "self": ArrowWeaveListType(types.Boolean()),
}
binary_input_type = {
    "self": ArrowWeaveListType(types.Boolean()),
    "other": types.UnionType(types.Boolean(), ArrowWeaveListType(types.Boolean())),
}

nullable_binary_input_type = {
    "self": ArrowWeaveListType(types.optional(types.Boolean())),
    "other": types.UnionType(
        types.optional(types.Boolean()),
        ArrowWeaveListType(types.optional(types.Boolean())),
    ),
}


null_consuming_binary_input_type_right = {
    "self": types.optional(types.Boolean()),
    "other": ArrowWeaveListType(types.optional(types.Boolean())),
}


self_type_output_type_fn = lambda input_types: input_types["self"]


@arrow_op(
    name="ArrowWeaveListBoolean-equal",
    input_type=nullable_binary_input_type,
    output_type=ArrowWeaveListType(types.Boolean()),
)
def bool_equal(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    result = util.equal(self._arrow_data, other)
    return ArrowWeaveList(result, types.Boolean(), self._artifact)


@arrow_op(
    name="ArrowWeaveListBoolean-notEqual",
    input_type=nullable_binary_input_type,
    output_type=ArrowWeaveListType(types.Boolean()),
)
def bool_not_equal(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    result = util.not_equal(self._arrow_data, other)
    return ArrowWeaveList(result, types.Boolean(), self._artifact)


@arrow_op(
    name="ArrowWeaveListBoolean-and",
    input_type=binary_input_type,
    output_type=self_type_output_type_fn,
)
def bool_and(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.and_(self._arrow_data, pc.fill_null(pc.cast(other, pa.bool_()), False)),
        types.Boolean(),
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListBoolean-or",
    input_type=binary_input_type,
    output_type=self_type_output_type_fn,
)
def bool_or(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.or_(self._arrow_data, pc.fill_null(pc.cast(other, pa.bool_()), False)),
        types.Boolean(),
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListBoolean-not",
    input_type=unary_input_type,
    output_type=self_type_output_type_fn,
)
def bool_not(self):
    return ArrowWeaveList(pc.invert(self._arrow_data), types.Boolean(), self._artifact)


def _cond_output_type(input_type):
    result_typed_dict = input_type["results"]
    if isinstance(result_typed_dict, ArrowWeaveListType):
        result_typed_dict = result_typed_dict.object_type
    return ArrowWeaveListType(
        types.optional(types.union(*result_typed_dict.property_types.values()))
    )


@arrow_op(
    name="ArrowWeaveList-cond",
    input_type={
        "cases": ArrowWeaveListType(types.TypedDict()),
        "results": types.union(
            ArrowWeaveListType(types.TypedDict()), types.TypedDict()
        ),
    },
    output_type=_cond_output_type,
)
def awl_cond(cases, results):
    cases = cases.without_tags()
    # Will crash if results are not of the same type.
    if isinstance(results, ArrowWeaveList):
        first_non_null_type = None
        null_fields = []
        results = results.without_tags()
        result_values = []
        for i in range(len(results._arrow_data.type)):
            field = results._arrow_data.field(i)
            if pa.types.is_null(field.type):
                null_fields.append(i)
            elif first_non_null_type is None:
                first_non_null_type = field.type
            result_values.append(field)
        result_typed_dict = results.object_type
        if first_non_null_type is not None:
            for i in null_fields:
                result_values[i] = result_values[i].cast(first_non_null_type)
    else:
        result_values = [pa.scalar(v) for v in results.values()]
        result_typed_dict = types.TypeRegistry.type_of(results)
    result_array = pc.case_when(cases._arrow_data, *result_values)
    result_object_type = types.optional(
        types.union(*result_typed_dict.property_types.values())
    )
    return ArrowWeaveList(result_array, result_object_type, cases._artifact)



from .. import context_state as _context_state

_loading_builtins_token = _context_state.set_loading_built_ins()

from . import panel_llm_monitor
from . import panel_trace_monitor
from . import panel_autoboard
from . import panel_seedboard
from . import panel_observability
from . import generator_templates

# This is just an example, uncomment to enable
# from . import panel_example_custom_board

_context_state.clear_loading_built_ins(_loading_builtins_token)



# Top-Level Namespace: weave.monitoring

# TODO: Notebook Walkthrough

from ..wandb_interface.wandb_stream_table import StreamTable
from .monitor import default_monitor, deinit_monitor, init_monitor

__all__ = ["StreamTable", "monitor", "openai"]



# Synthetic data for model monitoring

import datetime
import tqdm
import os
import numpy as np
import pandas as pd
import random
from faker import Faker
from weave.arrow.list_ import dataframe_to_arrow
from datetime import timedelta, time
from weave.arrow.list_ import ArrowWeaveList


# Generate the version schedule
def generate_version_schedule(
    start_date: datetime.datetime, end_date: datetime.datetime
) -> dict:
    current_date = start_date
    versions = ["1.0"]
    version_schedule = {}
    while current_date <= end_date:
        date_versions = []
        for version in versions:
            service_percent = random.uniform(0, 1)
            date_versions.append((version, service_percent))

        version_schedule[current_date.date()] = date_versions
        current_date += timedelta(days=1)
        if random.random() < 0.10:  # 5% chance to introduce a new version each day
            new_version = f"{float(versions[-1])+0.1:.1f}"
            versions.append(new_version)
        if len(versions) > 1 and random.random() < 0.10:
            versions.pop(0)
    return version_schedule


# Generate the latency schedule
def generate_latency_schedule(
    start_date: datetime.datetime, end_date: datetime.datetime
) -> dict:
    latency_schedule = {}
    for current_date in pd.date_range(start_date, end_date):
        base_latency = random.uniform(0.1, 1)
        day_factor = random.uniform(0.5, 1.5)
        month_factor = random.uniform(0.5, 1.5)
        latency = base_latency * day_factor * month_factor
        latency_schedule[current_date.date()] = latency
    return latency_schedule


# Generate the cost schedule
def generate_cost_schedule(
    start_date: datetime.datetime,
    end_date: datetime.datetime,
    cost_change_date: datetime.datetime,
) -> dict:
    cost_schedule = {}
    current_date = start_date
    cost_per_token = 0.01
    while current_date <= end_date:
        if current_date >= cost_change_date:
            cost_per_token = 0.005
        cost_schedule[current_date.date()] = cost_per_token
        current_date += timedelta(days=1)
    return cost_schedule


def generate_user_usage_schedule(
    start_date: datetime.datetime, end_date: datetime.datetime, users: list
) -> list:
    user_usage_schedule = []
    for user in users:
        current_date = start_date + timedelta(days=random.randrange(90))
        while current_date <= end_date:
            usage_periods = random.randint(1, 30)
            for _ in range(usage_periods):
                period_length_timedelta = timedelta(hours=random.randint(1, 24 * 7))
                rate = random.uniform(0.1, 10)
                user_usage_schedule.append(
                    (current_date, user, period_length_timedelta, rate)
                )
                current_date += period_length_timedelta  # Increment current_date
                if current_date > end_date:
                    break
    return user_usage_schedule


def random_predictions(n_users: int = 10) -> ArrowWeaveList:
    # Define our fake users
    fake = Faker()
    users = [fake.user_name() for _ in range(n_users)]

    # Read the file and generate prompts
    with open(
        os.path.abspath(
            os.path.join(os.path.dirname(__file__), "testdata/t8.shakespeare.txt")
        ),
        "r",
    ) as f:
        lines = f.read().split("\n")

    # Define the time range
    start_date = pd.to_datetime("2023-01-01", utc=True)
    end_date = pd.to_datetime("2023-03-31", utc=True)
    cost_change_date = pd.to_datetime("2023-02-15", utc=True)

    # Generate the schedules
    version_schedule = generate_version_schedule(start_date, end_date)
    latency_schedule = generate_latency_schedule(start_date, end_date)
    cost_schedule = generate_cost_schedule(start_date, end_date, cost_change_date)
    user_usage_schedule = generate_user_usage_schedule(start_date, end_date, users)

    # Helper function to generate a random completion
    def generate_completion(prompt: str) -> str:
        words = prompt.split()
        completion = " ".join(
            random.choices(words, k=int(len(words) * (random.random() + 0.1) * 10))
        )
        return completion

    data = []
    for usage in tqdm.tqdm(user_usage_schedule):
        usage_date, user, usage_period, rate = usage

        end_date = usage_date + usage_period
        increment = timedelta(hours=rate)

        while usage_date < end_date:
            # Find the version that was active during this usage
            if usage_date.date() not in version_schedule:
                break
            active_versions = version_schedule[usage_date.date()]
            # active_versions = [(version, percent) for date, version, percent in version_schedule if date.date() == usage_date.date()]
            # Normalize the service percentages
            total_percent = sum([percent for version, percent in active_versions])
            if total_percent == 0:
                continue
            normalized_percentages = [
                percent / total_percent for version, percent in active_versions
            ]

            version = np.random.choice(
                [v for v, p in active_versions], p=normalized_percentages
            )

            # Find the cost during this usage
            cost_per_token = cost_schedule[usage_date.date()]

            # Find the average latency during this usage
            latency = latency_schedule[usage_date.date()]
            latency *= 0.9 + random.random() * 0.2

            prompt = " ".join(random.sample(lines, 10))  # Increase prompt size
            completion = generate_completion(prompt)
            prompt_tokens = len(prompt.split())
            completion_tokens = len(completion.split())
            api_cost = (prompt_tokens + completion_tokens) * cost_per_token

            data.append(
                [
                    usage_date,
                    user,
                    version,
                    prompt,
                    completion,
                    prompt_tokens,
                    completion_tokens,
                    api_cost,
                    latency,
                ]
            )

            usage_date += increment

    df = pd.DataFrame(
        data,
        columns=[
            "timestamp",
            "username",
            "model_version",
            "prompt",
            "completion",
            "prompt_tokens",
            "completion_tokens",
            "api_cost",
            "latency",
        ],
    )
    return dataframe_to_arrow(df)  # type: ignore



import typing
from dataclasses import dataclass

from . import input_provider
from . import op_args
from . import op_def
from . import input_provider
from . import weave_types as types

# A GQLOutputTypeFn is a function that is called during the refinement phase of the compile pass
# to propagate the GQL keys of a node's input types to its output type.
GQLOutputTypeFn = typing.Callable[
    [input_provider.InputProvider, types.Type], types.Type
]


_ROOT_RESOLVERS: set[op_def.OpDef] = set()


@dataclass
class GqlOpPlugin:
    query_fn: typing.Callable[[input_provider.InputAndStitchProvider, str], str]
    is_root: bool = False
    root_resolver: typing.Optional["op_def.OpDef"] = None

    # given the input types to the op, return a new output type with the input types'
    # gql keys propagated appropriately. this is not a part of output_type to avoid
    # the UI needing to make additional network requests to get the output type
    gql_op_output_type: typing.Optional[GQLOutputTypeFn] = None

    def __post_init__(self) -> None:
        if self.root_resolver is not None:
            _ROOT_RESOLVERS.add(self.root_resolver)


# Ops in `domain_ops` can add this plugin to their op definition to indicate
# that they need data to be fetched from the GQL API. At it's core, the plugin
# allows the user to specify a `query_fn` that takes in the inputs to the op and
# returns a GQL query fragment that is needed by the calling op. The plugin also
# allows the user to specify whether the op is a root op which indicates it is
# the "top" of the GQL query tree. Note, while ops can use this directly (eg.
# see `project_ops.py::artifacts`), most ops use the higher level helpers
# defined in `wb_domain_gql.py`
def wb_gql_op_plugin(
    query_fn: typing.Callable[[input_provider.InputAndStitchProvider, str], str],
    is_root: bool = False,
    root_resolver: typing.Optional["op_def.OpDef"] = None,
    gql_op_output_type: typing.Optional[GQLOutputTypeFn] = None,
) -> dict[str, GqlOpPlugin]:
    return {
        "wb_domain_gql": GqlOpPlugin(
            query_fn, is_root, root_resolver, gql_op_output_type
        )
    }


def get_gql_plugin(
    op_def: "op_def.OpDef",
) -> typing.Optional[GqlOpPlugin]:
    if op_def.plugins is not None and "wb_domain_gql" in op_def.plugins:
        return op_def.plugins["wb_domain_gql"]
    return None


def first_arg_name(
    opdef: "op_def.OpDef",
) -> typing.Optional[str]:
    if not isinstance(opdef.input_type, op_args.OpNamedArgs):
        return None
    try:
        return next(iter(opdef.input_type.arg_types.keys()))
    except StopIteration:
        return None



import json
from ..gql_op_plugin import wb_gql_op_plugin
from ..api import op
from . import wb_domain_types as wdt
from ..language_features.tagging.make_tag_getter_op import make_tag_getter_op
from .wandb_domain_gql import (
    gql_prop_op,
    gql_direct_edge_op,
    gql_connection_op,
    gql_root_op,
)
from .. import weave_types as types

# Section 1/6: Tag Getters
#

# Section 2/6: Root Ops
#

# Section 3/6: Attribute Getters
gql_prop_op(
    "runQueue-id",
    wdt.RunQueueType,
    "id",
    types.String(),
)

# Section 5/6: Connection Ops
#

# Section 6/6: Non Standard Business Logic Ops
#



import dataclasses
import typing

import os
import weave

from . import chat_model


@weave.type()
class AnyscaleChatModel(chat_model.ChatModel):
    model_name: str = dataclasses.field(
        default_factory=lambda: "meta-llama/Llama-2-7b-chat-hf"
    )
    temperature: float = dataclasses.field(default_factory=lambda: 0.7)
    base_url: str = dataclasses.field(
        default_factory=lambda: "https://api.endpoints.anyscale.com/v1"
    )
    api_key_env_var: str = dataclasses.field(default_factory=lambda: "ANYSCALE_API_KEY")

    @weave.op()
    def complete(self, messages: typing.Any) -> typing.Any:
        from openai import OpenAI

        api_key = os.environ[self.api_key_env_var]

        base_url = self.base_url

        client = OpenAI(
            base_url=base_url,
            api_key=api_key,
            timeout=15,
        )

        from weave.monitoring.openai import patch

        patch()

        create = client.chat.completions.create

        response = create(
            model=self.model_name,
            messages=messages,
        )

        # TODO: return response when there is a weave type for ChatCompletion.
        return response.dict()



from typing import AsyncIterator, Callable, Iterable, Tuple, TypeVar, Awaitable
import asyncio

T = TypeVar("T")
U = TypeVar("U")


async def async_foreach(
    sequence: Iterable[T],
    func: Callable[[T], Awaitable[U]],
    max_concurrent_tasks: int,
) -> AsyncIterator[Tuple[T, U]]:
    semaphore = asyncio.Semaphore(max_concurrent_tasks)

    async def process_item(item: T) -> Tuple[T, U]:
        async with semaphore:
            result = await func(item)
            return item, result

    tasks = [asyncio.create_task(process_item(item)) for item in sequence]

    for task in asyncio.as_completed(tasks):
        item, result = await task
        yield item, result



import importlib
import logging
import os
import typing
from collections import defaultdict, Counter
from weave import context_state

logger = logging.getLogger(__name__)

loading_builtins_token = context_state.set_loading_built_ins()

ALL_MODULES = (
    "example",
    "bertviz",
    "xgboost",
    "shap",
    "sklearn",
    "torchvision",
    "torch_mnist_model_example",
    "huggingface",
    "craiyon",
    "spacy",
    "lens",
    "wandb",
    "scenario",
    "shawn",
    "replicate",
    "openai",
    # By declares a weave type for the python type "type" and causes problems
    # with op serialization
    # "py",
    "langchain",
    "umap",
    "hdbscan",
)

LOAD_RESULTS: typing.DefaultDict[str, list[str]] = defaultdict(list[str])

try:
    logger.info("Loading weave.ecosystem")
    for module in ALL_MODULES:
        try:
            globals()[module] = importlib.import_module(f"weave.ecosystem.{module}")
            LOAD_RESULTS["loaded"].append(module)
        except ImportError as exc:
            logger.debug(f"  {module} extension is unavailable, missing {exc.name}")
            LOAD_RESULTS["unavailable"].append(module)
        except Exception as exc:
            logger.error(f"  Error loading {module}")
            logger.error(exc)
            LOAD_RESULTS["error"].append(module)
    for result in ("loaded", "unavailable", "error"):
        modules = ", ".join(LOAD_RESULTS[result]) or "None"
        logger.info(f"  {result}: {modules}")

finally:
    context_state.clear_loading_built_ins(loading_builtins_token)



import typing

import pyarrow as pa
import pyarrow.compute as pc

from ..arrow.arrow import offsets_starting_at_zero

from ..api import op
from ..decorator_arrow_op import arrow_op
from .. import weave_types as types

from ..arrow.list_ import ArrowWeaveList, ArrowWeaveListType
from ..arrow.arrow import ArrowWeaveListType
from . import util


ARROW_WEAVE_LIST_STRING_TYPE = ArrowWeaveListType(types.String())
ARROW_WEAVE_LIST_BOOLEAN_TYPE = ArrowWeaveListType(types.Boolean())
ARROW_WEAVE_LIST_INT_TYPE = ArrowWeaveListType(types.Int())
ARROW_WEAVE_LIST_LIST_OF_STR_TYPE = ArrowWeaveListType(types.List(types.String()))

unary_input_type = {
    "self": ARROW_WEAVE_LIST_STRING_TYPE,
}
binary_input_type = {
    "self": ARROW_WEAVE_LIST_STRING_TYPE,
    "other": types.UnionType(
        types.optional(types.String()), ARROW_WEAVE_LIST_STRING_TYPE
    ),
}

null_consuming_binary_input_type = {
    "self": ArrowWeaveListType(types.optional(types.String())),
    "other": types.UnionType(
        types.optional(types.String()),
        ArrowWeaveListType(types.optional(types.String())),
    ),
}

null_consuming_binary_input_type_right = {
    "self": types.optional(types.String()),
    "other": ArrowWeaveListType(types.optional(types.String())),
}

self_type_output_type_fn = lambda input_types: input_types["self"]


def _concatenate_strings(
    left: ArrowWeaveList[str], right: typing.Union[str, ArrowWeaveList[str]]
) -> ArrowWeaveList[str]:
    a = left._arrow_data
    if right == None:
        return ArrowWeaveList(
            pa.nulls(len(a), type=a.type),
            types.NoneType(),
            left._artifact,
        )
    if isinstance(right, ArrowWeaveList):
        right = right._arrow_data
    return ArrowWeaveList(
        pc.binary_join_element_wise(left._arrow_data, right, ""),
        types.String(),
        left._artifact,
    )


@arrow_op(
    name="ArrowWeaveListString-equal",
    input_type=null_consuming_binary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def __eq__(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    result = util.equal(self._arrow_data, other)
    return ArrowWeaveList(result, types.Boolean(), self._artifact)


@arrow_op(
    name="ArrowWeaveListString-notEqual",
    input_type=null_consuming_binary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def __ne__(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    result = util.not_equal(self._arrow_data, other)
    return ArrowWeaveList(result, types.Boolean(), self._artifact)


@arrow_op(
    name="ArrowWeaveListString-contains",
    input_type=binary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def __contains__(self, other):
    if isinstance(other, ArrowWeaveList):
        return ArrowWeaveList(
            pa.array(
                other_item.as_py() in my_item.as_py()
                for my_item, other_item in zip(self._arrow_data, other._arrow_data)
            ),
            None,
            self._artifact,
        )
    return ArrowWeaveList(
        pc.match_substring(self._arrow_data, other), types.Boolean(), self._artifact
    )


# TODO: fix
@arrow_op(
    name="ArrowWeaveListString-in",
    input_type=binary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def in_(self, other):
    if isinstance(other, ArrowWeaveList):

        def new_val_iterator():
            for my_item, other_item in zip(self._arrow_data, other._arrow_data):
                if (
                    pa.compute.is_null(my_item).as_py()
                    or pa.compute.is_null(other_item).as_py()
                ):
                    yield None
                else:
                    yield my_item.as_py() in other_item.as_py()

        return ArrowWeaveList(
            pa.array(new_val_iterator()),
            types.Boolean(),
            self._artifact,
        )

    return ArrowWeaveList(
        # this has to be a python loop because the second argument to match_substring has to be a scalar string
        pa.array(
            item.as_py() in other if not pa.compute.is_null(item).as_py() else None
            for item in self._arrow_data
        ),
        types.Boolean(),
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListString-len",
    input_type=unary_input_type,
    output_type=ARROW_WEAVE_LIST_INT_TYPE,
)
def arrowweavelist_len(self):
    return ArrowWeaveList(
        pc.binary_length(self._arrow_data), types.Int(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListString-add",
    input_type=binary_input_type,
    output_type=self_type_output_type_fn,
)
def string_add(self, other):
    return _concatenate_strings(self, other)


# Handle plain str on the left. We'll need to figure out how to do
# this generally.
# TODO: Do this generally!
@op(
    name="ArrowWeaveListString_right-add",
    output_type=lambda input_type: input_type["right"],
)
def string_right_add(
    left: typing.Optional[str], right: ArrowWeaveList[typing.Optional[str]]
):
    return ArrowWeaveList(
        pc.binary_join_element_wise(left, right._arrow_data_asarray_no_tags(), ""),
        types.String(),
        right._artifact,
    )


@arrow_op(
    name="ArrowWeaveListString-append",
    input_type=binary_input_type,
    output_type=self_type_output_type_fn,
)
def append(self, other):
    return _concatenate_strings(self, other)


@arrow_op(
    name="ArrowWeaveListString-prepend",
    input_type=binary_input_type,
    output_type=self_type_output_type_fn,
)
def prepend(self, other):
    if isinstance(other, str):
        other = ArrowWeaveList(
            pa.array([other] * len(self._arrow_data)), types.String(), self._artifact
        )
    return _concatenate_strings(other, self)


@arrow_op(
    name="ArrowWeaveListString-split",
    input_type={
        "self": ARROW_WEAVE_LIST_STRING_TYPE,
        "pattern": types.UnionType(types.String(), ARROW_WEAVE_LIST_STRING_TYPE),
    },
    output_type=ARROW_WEAVE_LIST_LIST_OF_STR_TYPE,
)
def split(self, pattern):
    if isinstance(pattern, str):
        if isinstance(self._arrow_data, pa.DictionaryArray):
            # If we have a dictionary array, we need to split the dictionary and
            # then re-encode it as a dictionary array.
            return ArrowWeaveList(
                pc.split_pattern(self._arrow_data.dictionary, pattern),
                types.optional(types.List(types.String())),
                self._artifact,
            )
        return ArrowWeaveList(
            pc.split_pattern(self._arrow_data, pattern),
            types.optional(types.List(types.String())),
            self._artifact,
        )
    return ArrowWeaveList(
        pa.array(
            self._arrow_data[i].as_py().split(pattern._arrow_data[i].as_py())
            for i in range(len(self._arrow_data))
        ),
        types.optional(types.List(types.String())),
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListString-partition",
    input_type={
        "self": ARROW_WEAVE_LIST_STRING_TYPE,
        "sep": types.UnionType(types.String(), ARROW_WEAVE_LIST_STRING_TYPE),
    },
    output_type=ARROW_WEAVE_LIST_LIST_OF_STR_TYPE,
)
def partition(self, sep):
    def data_iterator():
        for i in range(len(self._arrow_data)):
            item = self._arrow_data[i].as_py()
            separator = sep if isinstance(sep, str) else sep._arrow_data[i].as_py()
            if not (item is None or separator is None):
                yield item.partition(separator)
            else:
                yield None

    return ArrowWeaveList(
        pa.array(data_iterator()),
        types.optional(types.List(types.String())),
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListString-startsWith",
    input_type={
        "self": ARROW_WEAVE_LIST_STRING_TYPE,
        "prefix": types.UnionType(types.String(), ARROW_WEAVE_LIST_STRING_TYPE),
    },
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def startswith(self, prefix):
    if isinstance(prefix, str):
        return ArrowWeaveList(
            pc.starts_with(self._arrow_data, prefix), types.Boolean(), self._artifact
        )
    return ArrowWeaveList(
        pa.array(
            s.as_py().startswith(p.as_py())
            for s, p in zip(self._arrow_data, prefix._arrow_data)
        ),
        types.Boolean(),
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListString-endsWith",
    input_type={
        "self": ARROW_WEAVE_LIST_STRING_TYPE,
        "suffix": types.UnionType(types.String(), ARROW_WEAVE_LIST_STRING_TYPE),
    },
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def endswith(self, suffix):
    if isinstance(suffix, str):
        return ArrowWeaveList(
            pc.ends_with(self._arrow_data, suffix), types.Boolean(), self._artifact
        )
    return ArrowWeaveList(
        pa.array(
            s.as_py().endswith(p.as_py())
            for s, p in zip(self._arrow_data, suffix._arrow_data)
        ),
        types.Boolean(),
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListString-isAlpha",
    input_type=unary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def isalpha(self):
    return ArrowWeaveList(
        pc.ascii_is_alpha(self._arrow_data), types.Boolean(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListString-isNumeric",
    input_type=unary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def isnumeric(self):
    return ArrowWeaveList(
        pc.ascii_is_decimal(self._arrow_data), types.Boolean(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListString-isAlnum",
    input_type=unary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def isalnum(self):
    return ArrowWeaveList(
        pc.ascii_is_alnum(self._arrow_data), types.Boolean(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListString-lower",
    input_type=unary_input_type,
    output_type=ArrowWeaveListType(types.String()),
)
def lower(self):
    return ArrowWeaveList(
        pc.ascii_lower(self._arrow_data), types.String(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListString-upper",
    input_type=unary_input_type,
    output_type=ArrowWeaveListType(types.String()),
)
def upper(self):
    return ArrowWeaveList(
        pc.ascii_upper(self._arrow_data), types.String(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListString-slice",
    input_type={
        "self": ARROW_WEAVE_LIST_STRING_TYPE,
        "begin": types.Int(),
        "end": types.Int(),
    },
    output_type=self_type_output_type_fn,
)
def slice(self, begin, end):
    return ArrowWeaveList(
        pc.utf8_slice_codeunits(self._arrow_data, begin, end),
        types.String(),
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListString-replace",
    input_type={
        "self": ARROW_WEAVE_LIST_STRING_TYPE,
        "pattern": types.String(),
        "replacement": types.String(),
    },
    output_type=self_type_output_type_fn,
)
def replace(self, pattern, replacement):
    return ArrowWeaveList(
        pc.replace_substring(self._arrow_data, pattern, replacement),
        types.String(),
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListString-strip",
    input_type=unary_input_type,
    output_type=self_type_output_type_fn,
)
def strip(self):
    return ArrowWeaveList(
        pc.utf8_trim_whitespace(self._arrow_data), types.String(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListString-lStrip",
    input_type=unary_input_type,
    output_type=self_type_output_type_fn,
)
def lstrip(self):
    return ArrowWeaveList(
        pc.utf8_ltrim_whitespace(self._arrow_data), types.String(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListString-rStrip",
    input_type=unary_input_type,
    output_type=self_type_output_type_fn,
)
def rstrip(self):
    return ArrowWeaveList(
        pc.utf8_rtrim_whitespace(self._arrow_data), types.String(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListString-joinToStr",
    input_type={
        "arr": ArrowWeaveListType(types.List(types.optional(types.String()))),
        "sep": types.UnionType(types.String(), ArrowWeaveListType(types.String())),
    },
    output_type=ArrowWeaveListType(types.String()),
)
def join_to_str(arr, sep):
    if isinstance(sep, ArrowWeaveList):
        sep = sep._arrow_data
    # match Weave0 - join nulls to empty string
    filled_arr = pa.ListArray.from_arrays(
        offsets_starting_at_zero(arr._arrow_data),
        arr._arrow_data.flatten().fill_null(""),
    )
    return ArrowWeaveList(
        pc.binary_join(filled_arr, sep), types.String(), arr._artifact
    )


@arrow_op(
    name="ArrowWeaveListString-toNumber",
    input_type={"self": ArrowWeaveListType(types.String())},
    output_type=ArrowWeaveListType(types.Number()),
)
def to_number(self):
    arrow_data = self._arrow_data
    is_not_numeric = pc.invert(pc.utf8_is_numeric(arrow_data))

    # need to use kleene logic here for masking
    mask = pc.and_kleene(is_not_numeric, pa.nulls(len(self)).cast(pa.bool_()))

    new_data = pc.replace_with_mask(arrow_data, mask, arrow_data).cast(pa.float64())
    return ArrowWeaveList(new_data, types.optional(types.Number()), self._artifact)



import copy
import dataclasses
import functools
import hashlib
import uuid
import time
import json
import typing
from typing import Sequence

from collections.abc import Mapping
from .graph_client import GraphClient
from . import urls
from . import context_state
from . import weave_internal
from . import monitoring
from .monitoring import monitor
from . import artifact_wandb
from . import op_def
from .ref_base import Ref
from . import stream_data_interfaces
from .eager import WeaveIter, select_all
from .run import RunKey, Run
from .run_streamtable_span import RunStreamTableSpan
from . import stream_data_interfaces


def refs_to_str(val: typing.Any) -> typing.Any:
    if isinstance(val, Ref):
        return str(val)
    elif isinstance(val, dict):
        return {k: refs_to_str(v) for k, v in val.items()}  # type: ignore
    elif isinstance(val, list):
        return [refs_to_str(v) for v in val]  # type: ignore
    else:
        return val


def hash_inputs(
    inputs: Mapping[str, typing.Any],
) -> str:
    hasher = hashlib.md5()
    hasher.update(json.dumps(refs_to_str(inputs)).encode())
    return hasher.hexdigest()


class GraphClientWandbArtStreamTable(GraphClient[RunStreamTableSpan]):
    entity_name: str
    project_name: str

    def __init__(self, entity_name: str, project_name: str):
        self.entity_name = entity_name
        self.project_name = project_name

    @functools.cached_property
    def runs_st(self) -> monitoring.StreamTable:
        return monitoring.StreamTable(f"{self.entity_name}/{self.project_name}/stream")

    @functools.cached_property
    def run_feedback_st(self) -> monitoring.StreamTable:
        return monitoring.StreamTable(
            f"{self.entity_name}/{self.project_name}/run-feedback"
        )

    ##### Read API

    def runs(self) -> WeaveIter[RunStreamTableSpan]:
        return WeaveIter(self.runs_st.rows(), cls=RunStreamTableSpan)

    def run(self, run_id: str) -> typing.Optional[RunStreamTableSpan]:
        with context_state.lazy_execution():
            rows_node = self.runs_st.rows()
            filter_node = rows_node.filter(lambda row: row["span_id"] == run_id)[0]  # type: ignore
            run_attrs = weave_internal.use(select_all(filter_node))
            if not isinstance(run_attrs, dict):
                return None
            if run_attrs["span_id"] == None:
                return None
            run_attrs = typing.cast(stream_data_interfaces.TraceSpanDict, run_attrs)
            return RunStreamTableSpan(run_attrs)

    def find_op_run(
        self, op_name: str, inputs: dict[str, typing.Any]
    ) -> typing.Optional[RunStreamTableSpan]:
        from . import ops_primitives

        inputs_digest = hash_inputs(inputs)
        from . import compile

        with context_state.lazy_execution():
            with compile.enable_compile():
                rows_node = self.runs_st.rows()
                filter_node = rows_node.filter(  # type: ignore
                    lambda row: ops_primitives.Boolean.bool_and(
                        row["name"] == op_name,
                        row["attributes"]["_inputs_digest"] == inputs_digest,
                    )
                )[
                    0
                ]  # type: ignore
                run_attrs = weave_internal.use(select_all(filter_node))
                if not isinstance(run_attrs, dict):
                    return None
                if run_attrs.get("span_id") == None:
                    return None
                run_attrs = typing.cast(stream_data_interfaces.TraceSpanDict, run_attrs)
                return RunStreamTableSpan(run_attrs)

    def run_children(self, run_id: str) -> WeaveIter[RunStreamTableSpan]:
        with context_state.lazy_execution():
            rows_node = self.runs_st.rows()
            filter_node = rows_node.filter(lambda row: row["parent_id"] == run_id)  # type: ignore
            return WeaveIter(filter_node, cls=RunStreamTableSpan)

    # Hmm... I want this to be a ref to an op I think?
    def op_runs(self, op_def: op_def.OpDef) -> WeaveIter[RunStreamTableSpan]:
        with context_state.lazy_execution():
            rows_node = self.runs_st.rows()
            filter_node = rows_node.filter(  # type: ignore
                lambda row: row["name"] == str(op_def.location)
            )
            return WeaveIter(filter_node, cls=RunStreamTableSpan)

    def ref_input_to(self, ref: Ref) -> WeaveIter[RunStreamTableSpan]:
        with context_state.lazy_execution():
            rows_node = self.runs_st.rows()
            filter_node = rows_node.filter(lambda row: row["inputs._ref0"] == ref)  # type: ignore
            return WeaveIter(filter_node, cls=RunStreamTableSpan)

    def ref_value_input_to(self, ref: Ref) -> WeaveIter[RunStreamTableSpan]:
        with context_state.lazy_execution():
            rows_node = self.runs_st.rows()
            filter_node = rows_node.filter(lambda row: row["inputs._ref_digest0"] == ref.digest)  # type: ignore
            return WeaveIter(filter_node, cls=RunStreamTableSpan)

    def ref_output_of(self, ref: Ref) -> typing.Optional[RunStreamTableSpan]:
        with context_state.lazy_execution():
            rows_node = self.runs_st.rows()
            filter_node = rows_node.filter(lambda row: row["outputs._ref0"] == ref)[0]  # type: ignore
            run_attrs = weave_internal.use(select_all(filter_node))
            if not isinstance(run_attrs, dict):
                return None
            if run_attrs["span_id"] == None:
                return None
            run_attrs = typing.cast(stream_data_interfaces.TraceSpanDict, run_attrs)
            return RunStreamTableSpan(run_attrs)

    def run_feedback(self, run_id: str) -> WeaveIter[dict[str, typing.Any]]:
        with context_state.lazy_execution():
            rows_node = self.run_feedback_st.rows()
            filter_node = rows_node.filter(lambda row: row["run_id"] == run_id)  # type: ignore
            return WeaveIter(filter_node)

    def feedback(self, feedback_id: str) -> typing.Optional[dict[str, typing.Any]]:
        with context_state.lazy_execution():
            rows_node = self.run_feedback_st.rows()
            filter_node = rows_node.filter(lambda row: row["feedback_id"] == feedback_id)[0]  # type: ignore
            feedback_attrs = weave_internal.use(select_all(filter_node))
            if not isinstance(feedback_attrs, dict):
                return None
            if feedback_attrs["feedback_id"] == None:
                return None
            return feedback_attrs

    # Helpers

    def ref_is_own(self, ref: typing.Optional[Ref]) -> bool:
        return isinstance(ref, artifact_wandb.WandbArtifactRef)

    def ref_uri(
        self, name: str, version: str, path: str
    ) -> artifact_wandb.WeaveWBArtifactURI:
        return artifact_wandb.WeaveWBArtifactURI(
            name, version, self.entity_name, self.project_name, path=path
        )

    def run_ui_url(self, run: Run) -> str:
        return urls.call_path_as_peek(
            self.entity_name,
            self.project_name,
            run.id,
        )

    ##### Write API

    def save_object(
        self, obj: typing.Any, name: str, branch_name: str
    ) -> artifact_wandb.WandbArtifactRef:
        from . import storage

        res = storage._direct_publish(
            obj,
            name=name,
            wb_entity_name=self.entity_name,
            wb_project_name=self.project_name,
            branch_name=branch_name,
        )
        return res

    def create_run(
        self,
        op_name: str,
        parent: typing.Optional["RunKey"],
        inputs: typing.Dict[str, typing.Any],
        input_refs: Sequence[Ref],
    ) -> RunStreamTableSpan:
        inputs_digest = hash_inputs(inputs)
        attrs = {"_inputs_digest": inputs_digest}
        inputs = copy.copy(inputs)
        inputs["_keys"] = list(inputs.keys())
        for i, ref in enumerate(input_refs[:3]):
            inputs["_ref%s" % i] = ref
            inputs["_ref_digest%s" % i] = ref.digest

        if parent:
            trace_id = parent.trace_id
            parent_id = parent.id
        else:
            trace_id = str(uuid.uuid4())
            parent_id = None
        cur_time = time.time()
        span = stream_data_interfaces.TraceSpanDict(
            span_id=str(uuid.uuid4()),
            trace_id=trace_id,
            parent_id=parent_id,
            name=op_name,
            status_code="UNSET",
            start_time_s=cur_time,
            end_time_s=cur_time,  # currently required, so set to start time for now?
            inputs=inputs,
            attributes=attrs,
            output=None,
            summary=None,
            exception=None,
        )
        # Don't log create for now
        # self.runs_st.log(span)
        return RunStreamTableSpan(span)

    def fail_run(self, run: RunStreamTableSpan, exception: BaseException) -> None:
        span = copy.copy(run._attrs)
        span["end_time_s"] = time.time()
        span["status_code"] = "ERROR"
        span["exception"] = str(exception)
        span["summary"] = {"latency_s": span["end_time_s"] - span["start_time_s"]}
        self.runs_st.log(span)

    def finish_run(
        self,
        run: RunStreamTableSpan,
        output: typing.Any,
        output_refs: Sequence[Ref],
    ) -> None:
        span = copy.copy(run._attrs)
        output = copy.copy(output)
        if not isinstance(output, dict):
            output = {"_result": output}
        output["_keys"] = list(output.keys())
        for i, ref in enumerate(output_refs[:3]):
            output["_ref%s" % i] = ref
            output["_ref_digest%s" % i] = ref.digest
        span["end_time_s"] = time.time()
        span["status_code"] = "SUCCESS"
        span["output"] = output
        span["summary"] = {"latency_s": span["end_time_s"] - span["start_time_s"]}
        span["attributes"] = monitor._attributes.get()

        self.runs_st.log(span)

    def add_feedback(self, run_id: str, feedback: dict[str, typing.Any]) -> None:
        feedback_id = str(uuid.uuid4())
        self.run_feedback_st.log(
            {"run_id": run_id, "feedback_id": feedback_id, "feedback": feedback}
        )



import weave
from .. import weave_internal as internal
from .. import weave_types as types
from .generator_templates import template_registry


# IMPORTANT: To enable this (or any template), it must be imported in
# weave/panels_py/__init__.py This example is explicitly commented out since it
# is not intended to be in production. however, you are encouraged to uncomment
# it and play around with it.

# BOARD_ID must be unique across all ops. It must only contain letters and underscores.
BOARD_ID = "example_custom_board"

# BOARD_DISPLAY_NAME is the name that will be displayed in the UI
BOARD_DISPLAY_NAME = "Example Custom Board"

# BOARD_DESCRIPTION is the description that will be displayed in the UI
BOARD_DESCRIPTION = "An example board for internal developers"

# BOARD_INPUT_WEAVE_TYPE is the weave type of the input node.
BOARD_INPUT_WEAVE_TYPE = types.List(
    types.TypedDict({})
    # Example of requiring an `api_cost` field
    # types.TypedDict({"api_cost": types.optional(weave.types.Number())})
)


board_name = "py_board-" + BOARD_ID


@weave.op(  # type: ignore
    name=board_name,
    hidden=True,
    input_type={
        "input_node": types.Function(
            {},
            BOARD_INPUT_WEAVE_TYPE,
        )
    },
)
def board(
    input_node,
) -> weave.panels.Group:
    # Define your VarBar variables here
    control_items = [
        weave.panels.GroupPanel(internal.const("Example Custom Board"), id="title"),
        weave.panels.GroupPanel(input_node, id="data"),
    ]

    title_var = internal.make_var_node(input_node.type, "title")
    data_var = internal.make_var_node(input_node.type, "data")

    # Define your panels here
    panels = [
        weave.panels.BoardPanel(title_var, id="title_panel"),
        weave.panels.BoardPanel(data_var, id="table"),
    ]

    # Example of using the input type to modify the board:
    # Here, we add a panel for each numeric column, showing a distribution
    for column_name, column_type in input_node.type.object_type.property_types.items():
        if types.optional(types.Number()).assign_type(column_type):
            panels.append(
                weave.panels.BoardPanel(
                    data_var[column_name],  # type: ignore
                    id=column_name + "_distribution",
                )
            )

    return weave.panels.Board(vars=control_items, panels=panels)


template_registry.register(
    board_name,
    BOARD_DISPLAY_NAME,
    BOARD_DESCRIPTION,
)



# Concat implementation for ArrowWeaveList. Should work in all cases!
#
# Any places where we extend lists need to follow our merge_types implementation.
# Fortunately, concat is the only place that extends lists! Everything else should
# make use of concat to do so.


import typing
import dataclasses
import numpy as np
import pyarrow as pa
from pyarrow import compute as pc

from .. import weave_types as types
from ..language_features.tagging import tagged_value_type

from .. import errors
from .list_ import (
    ArrowWeaveList,
    ArrowWeaveListGeneric,
    is_typedict_arrowweavelist,
    is_object_arrowweavelist,
    is_taggedvalue_arrowweavelist,
    is_list_arrowweavelist,
    is_ref_arrowweavelist,
    unsafe_awl_construction,
    offsets_starting_at_zero,
    make_vec_none,
)

DEBUG = False

from .. import engine_trace

tracer = engine_trace.tracer()


@dataclasses.dataclass
class UnionMember:
    values: ArrowWeaveList
    mask: pa.Array
    offsets: pa.Array


def indent_print(indent: int = 0, *args) -> None:
    if DEBUG:
        print("    " * indent + " ".join([str(a) for a in args]))


def pa_concat_arrays(arrays: list[typing.Union[pa.Array, pa.ChunkedArray]]) -> pa.Array:
    arrays = [a if isinstance(a, pa.Array) else a.combine_chunks() for a in arrays]

    with tracer.trace("pa_concat_arrays"):
        try:
            return pa.concat_arrays(arrays)
        except pa.lib.ArrowNotImplementedError as e:
            # pyarrow doesn't support concatenating all dictionaries, so we have to do it ourselves.
            if all(isinstance(a, pa.DictionaryArray) for a in arrays):
                # We can concatenate dictionaries by concatenating the indices and the dictionary
                # values separately.

                index_offset = 0
                indices = []
                for a in arrays:
                    indices.append(pc.add(a.indices, index_offset).cast(pa.int32()))
                    index_offset += len(a.dictionary)

                indices = pa_concat_arrays(indices)
                dictionary = pa_concat_arrays([a.dictionary for a in arrays])
                return pa.DictionaryArray.from_arrays(indices, dictionary)
            else:
                raise


def _concatenate_typeddicts(
    l1: ArrowWeaveListGeneric[types.TypedDict],
    l2: ArrowWeaveListGeneric[types.TypedDict],
    depth: int = 0,
) -> ArrowWeaveList:
    # Columns are in the order they're encountered.
    all_keys = {**l1.object_type.property_types, **l2.object_type.property_types}
    if not all_keys:
        # Both are empty
        return ArrowWeaveList(
            pa_concat_arrays([l1._arrow_data, l2._arrow_data]),
            types.TypedDict({}),
            invalid_reason="Possibly nullable",
        )
    properties: dict[str, ArrowWeaveList] = {}
    for key in all_keys:
        indent_print(depth, "TypedDict key", key)
        if key not in l1.object_type.property_types:
            properties[key] = _concatenate(
                make_vec_none(len(l1._arrow_data)), l2.column(key), depth=depth + 1
            )
        elif key not in l2.object_type.property_types:
            properties[key] = _concatenate(
                l1.column(key), make_vec_none(len(l2)), depth=depth + 1
            )
        else:
            properties[key] = _concatenate(
                l1.column(key), l2.column(key), depth=depth + 1
            )

    return ArrowWeaveList(
        pa.StructArray.from_arrays(
            [properties[key]._arrow_data for key in all_keys],
            list(all_keys),
            mask=pa_concat_arrays(
                [
                    pa.compute.is_null(l1._arrow_data),
                    pa.compute.is_null(l2._arrow_data),
                ]
            ),
        ),
        types.TypedDict({key: properties[key].object_type for key in all_keys}),
        l1._artifact,
        invalid_reason="Possibly nullable",
    )


def _concatenate_objects(
    l1: ArrowWeaveListGeneric[types.ObjectType],
    l2: ArrowWeaveListGeneric[types.ObjectType],
    depth: int = 0,
) -> ArrowWeaveList:
    assert l1.object_type.name == l2.object_type.name
    attrs: dict[str, ArrowWeaveList] = {}
    merged_type = typing.cast(
        types.ObjectType, types.merge_types(l1.object_type, l2.object_type)
    )
    for key in merged_type.property_types():
        attrs[key] = _concatenate(l1.column(key), l2.column(key), depth=depth + 1)
    return ArrowWeaveList(
        pa.StructArray.from_arrays(
            [attrs[key]._arrow_data for key in attrs],
            list(attrs.keys()),
            mask=pa_concat_arrays(
                [
                    pa.compute.is_null(l1._arrow_data),
                    pa.compute.is_null(l2._arrow_data),
                ]
            ),
        ),
        merged_type,
        l1._artifact,
        invalid_reason="Possibly nullable",
    )


def _concatenate_taggedvalues(
    l1: ArrowWeaveListGeneric[tagged_value_type.TaggedValueType],
    l2: ArrowWeaveListGeneric[tagged_value_type.TaggedValueType],
    depth: int = 0,
) -> ArrowWeaveList:
    tag = _concatenate(l1.tagged_value_tag(), l2.tagged_value_tag(), depth=depth + 1)
    value = _concatenate(
        l1.tagged_value_value(), l2.tagged_value_value(), depth=depth + 1
    )
    return ArrowWeaveList(
        pa.StructArray.from_arrays(
            [tag._arrow_data, value._arrow_data],
            ["_tag", "_value"],
            mask=pa_concat_arrays(
                [
                    pa.compute.is_null(l1._arrow_data),
                    pa.compute.is_null(l2._arrow_data),
                ]
            ),
        ),
        tagged_value_type.TaggedValueType(tag.object_type, value.object_type),  # type: ignore
        l1._artifact,
        invalid_reason="Possibly nullable",
    )


def convert_null_array_to_list(
    array: typing.Union[pa.ListArray, pa.NullArray]
) -> pa.ListArray:
    if isinstance(array, pa.NullArray):
        return pc.cast(array, pa.list_(pa.null()))
    else:
        return array


def _concatenate_lists(
    l1: ArrowWeaveListGeneric[types.List],
    l2: ArrowWeaveListGeneric[types.List],
    depth: int = 0,
) -> ArrowWeaveList:
    l1_arrow_data = convert_null_array_to_list(l1._arrow_data)
    l2_arrow_data = convert_null_array_to_list(l2._arrow_data)

    self_values: ArrowWeaveList = ArrowWeaveList(
        l1_arrow_data.flatten(), l1.object_type.object_type, l2._artifact
    )
    other_values: ArrowWeaveList = ArrowWeaveList(
        l2_arrow_data.flatten(), l2.object_type.object_type, l2._artifact
    )
    concatted_values = _concatenate(self_values, other_values, depth=depth + 1)
    new_offsets = pa_concat_arrays(
        [
            offsets_starting_at_zero(l1_arrow_data)[:-1],
            pa.compute.add(
                offsets_starting_at_zero(l2_arrow_data),
                len(self_values),
            ).cast(pa.int32()),
        ]
    )
    return ArrowWeaveList(
        pa.ListArray.from_arrays(
            new_offsets,
            concatted_values._arrow_data,
            mask=pa_concat_arrays(
                [
                    pa.compute.is_null(l1_arrow_data),
                    pa.compute.is_null(l2_arrow_data),
                ]
            ),
        ),
        types.List(
            types.merge_types(l1.object_type.object_type, l2.object_type.object_type)
        ),
        l1._artifact,
        invalid_reason="Possibly nullable",
    )


def _concatenate_non_unions(
    self: ArrowWeaveList, other: ArrowWeaveList, depth: int = 0
) -> typing.Optional[ArrowWeaveList]:
    if is_typedict_arrowweavelist(self) and is_typedict_arrowweavelist(other):
        return _concatenate_typeddicts(self, other, depth=depth)
    elif is_object_arrowweavelist(self) and is_object_arrowweavelist(other):
        if self.object_type.name == other.object_type.name:
            return _concatenate_objects(self, other, depth=depth)
    elif is_taggedvalue_arrowweavelist(self) and is_taggedvalue_arrowweavelist(other):
        return _concatenate_taggedvalues(self, other, depth=depth)
    elif is_list_arrowweavelist(self) and is_list_arrowweavelist(other):
        # Lists with different object types
        return _concatenate_lists(self, other, depth=depth)

    if isinstance(self.object_type, types.Number) and isinstance(
        other.object_type, types.Number
    ):
        # Numeric special case
        indent_print(
            depth, "Extend case number types", self.object_type, other.object_type
        )

        # Weave numeric types are odd for legacy reasons. We have three numeric types
        # Number, Float, Int. Unfortunately Number is a Union of Float and Int, but
        # also assignable to Int.

        # We should fix that. But for now we rely on the arrow type for this special
        # case.

        if self._arrow_data.type == other._arrow_data.type:
            new_arrow_type = self._arrow_data.type
        else:
            # If types aren't equal, just cast to float64. This is not entirely
            # correct! Very large integers will be truncated.
            new_arrow_type = pa.float64()

        new_weave_type: types.Type
        if pa.types.is_integer(new_arrow_type):
            new_weave_type = types.Int()
        elif pa.types.is_floating(new_arrow_type):
            new_weave_type = types.Float()
        else:
            raise errors.WeaveInternalError('Unexpected type for "Number" type.')

        if len(self) == 0:
            data = other._arrow_data
        elif len(other) == 0:
            data = self._arrow_data
        else:
            data = pa_concat_arrays(
                [
                    self._arrow_data.cast(new_arrow_type),
                    other._arrow_data.cast(new_arrow_type),
                ]
            )
        return ArrowWeaveList(
            data, new_weave_type, self._artifact, invalid_reason="Possibly nullable"
        )
    elif (
        # Types exactly equal case
        self.object_type == other.object_type
        or (
            # Types are refs, so arrow type is string
            is_ref_arrowweavelist(self)
            and is_ref_arrowweavelist(other)
            and self.object_type.__class__ == other.object_type.__class__
        )
    ):
        indent_print(depth, "Extend case equal", self.object_type, other.object_type)
        if len(self) == 0:
            data = other._arrow_data
        elif len(other) == 0:
            data = self._arrow_data
        else:
            data = pa_concat_arrays([self._arrow_data, other._arrow_data])
        return ArrowWeaveList(
            data,
            types.merge_types(self.object_type, other.object_type),
            self._artifact,
            invalid_reason="Possibly nullable",
        )
    return None


def _concatenate(
    self: "ArrowWeaveList", other: "ArrowWeaveList", depth=0
) -> "ArrowWeaveList":
    if depth > 50:
        raise ValueError("Maximum recursion depth exceeded")
    indent_print(depth, "EXTEND TOP", depth, self.object_type, other.object_type)

    # We use UnknownType for two things:
    #   1. to indicate that we don't know the type of an object,
    #   2. as the object type of a zero length array.
    # The former case, where we don't know the type of an object,
    # should never happen here! This is runtime and we know the types of
    # things.

    # Zero length arrays of unknown type
    if self.object_type == types.UnknownType():
        if len(self) == 0:
            return other
        else:
            raise errors.WeaveInternalError(
                'Encountered non-zero length "UnknownType" array'
            )
    if other.object_type == types.UnknownType():
        if len(other) == 0:
            return self
        else:
            raise errors.WeaveInternalError(
                'Encountered non-zero length "UnknownType" array'
            )

    # Cases where one of the types is None
    if isinstance(self.object_type, types.NoneType) and isinstance(
        other.object_type, types.NoneType
    ):
        return make_vec_none(len(self) + len(other))
    elif isinstance(self.object_type, types.NoneType):
        indent_print(depth, "Extend case self None")
        if len(self) == 0:
            return other
        elif len(other) == 0:
            return self
        self_data = pa.nulls(len(self), other._arrow_data.type)
        other_data = other._arrow_data
        data = pa_concat_arrays([self_data, other_data])
        return ArrowWeaveList(data, types.optional(other.object_type), other._artifact)
    elif isinstance(other.object_type, types.NoneType):
        indent_print(depth, "Extend case other None")
        if len(self) == 0:
            return other
        elif len(other) == 0:
            return self
        self_data = self._arrow_data
        other_data = pa.nulls(len(other), self._arrow_data.type)
        data = pa_concat_arrays([self_data, other_data])
        return ArrowWeaveList(data, types.optional(self.object_type), self._artifact)

    # Separate nulls from type
    self_nullable, self_non_none_type = types.split_none(self.object_type)
    other_nullable, other_non_none_type = types.split_none(other.object_type)
    self = self._with_object_type(
        self_non_none_type, invalid_reason="Possibly nullable"
    )
    other = other._with_object_type(
        other_non_none_type, invalid_reason="Possibly nullable"
    )
    result_nullable = self_nullable or other_nullable

    if not isinstance(self.object_type, types.UnionType) and not isinstance(
        other.object_type, types.UnionType
    ):
        result = _concatenate_non_unions(self, other, depth=depth)
        if result is not None:
            if result_nullable:
                return result._with_object_type(types.optional(result.object_type))
            else:
                return result._clear_invalid_reason()

    # Otherwise, we have types that can't be merged, we're going to produce a union

    if isinstance(self.object_type, types.UnionType):
        self_field_types = self.object_type.members
        self_type_codes = self._arrow_data.type_codes
        self_offsets = self._arrow_data.offsets
        self_fields = [
            self._arrow_data.field(i) for i in range(len(self._arrow_data.type))
        ]
    else:
        self_field_types = [self.object_type]
        self_type_codes = pa.array(np.zeros(len(self), dtype=np.int8))
        self_offsets = pa.array(np.arange(len(self), dtype=np.int32))
        self_fields = [self._arrow_data]

    if isinstance(other.object_type, types.UnionType):
        other_field_types = other.object_type.members
        other_type_codes = other._arrow_data.type_codes
        other_offsets = other._arrow_data.offsets
        other_fields = [
            other._arrow_data.field(i) for i in range(len(other._arrow_data.type))
        ]
    else:
        other_field_types = [other.object_type]
        other_type_codes = pa.array(np.zeros(len(other), dtype=np.int8))
        other_offsets = pa.array(np.arange(len(other), dtype=np.int32))
        other_fields = [other._arrow_data]

    assert len(self_field_types) == len(self_fields)
    assert len(other_field_types) == len(other_fields)

    # produce merged union
    new_member_awls: list[UnionMember] = []

    indent_print(depth, "MERGING UNIONS")
    indent_print(depth, "SELF NULLABLE", self_nullable)
    # indent_print(depth, "SELF MEMBERS")
    # for member in self_fields:
    #     indent_print(depth + 1, member)
    indent_print(depth, "OTHER NULLABLE", other_nullable)
    # indent_print(depth, "OTHER MEMBERS")
    # for member in other_fields:
    #     indent_print(depth + 1, member)

    remaining_other_indexes = {i: True for i in range(len(other_fields))}
    for self_i, (self_member_type, self_field) in enumerate(
        zip(self_field_types, self_fields)
    ):
        for other_i in remaining_other_indexes:
            other_member_type = other_field_types[other_i]
            other_field = other_fields[other_i]
            if types.types_are_mergeable(self_member_type, other_member_type):
                indent_print(
                    depth,
                    "SELF OTHER FIELD merge",
                    self_i,
                    other_i,
                    "self type:",
                    self_member_type,
                    "other type:",
                    other_member_type,
                )
                concatted = _concatenate_non_unions(
                    ArrowWeaveList(
                        self_field, self_member_type, invalid_reason="Possibly nullable"
                    ),
                    ArrowWeaveList(
                        other_field,
                        other_member_type,
                        invalid_reason="Possibly nullable",
                    ),
                )
                if concatted is None:
                    raise errors.WeaveInternalError(
                        "Could not concatenate mergable types"
                    )
                new_member_awls.append(
                    UnionMember(
                        values=concatted,
                        mask=pa_concat_arrays(
                            [
                                pa.compute.equal(self_type_codes, self_i).cast(
                                    pa.int8()
                                ),
                                pa.compute.equal(other_type_codes, other_i).cast(
                                    pa.int8()
                                ),
                            ]
                        ),
                        offsets=pa_concat_arrays(
                            [
                                self_offsets,
                                pa.compute.add(other_offsets, len(self_field)).cast(
                                    pa.int32()
                                ),
                            ]
                        ),
                    )
                )
                indent_print(
                    depth,
                    "SELF OTHER FIELD merge RESULT",
                    new_member_awls[-1].values.object_type,
                )
                remaining_other_indexes.pop(other_i)
                break
        else:
            # no matching other member found
            indent_print(depth, "SELF FIELD solo", self_i, self_member_type)
            new_member_awls.append(
                UnionMember(
                    values=ArrowWeaveList(
                        self_fields[self_i],
                        self_member_type,
                        self._artifact,
                        invalid_reason="Possibly nullable",
                    ),
                    mask=pa_concat_arrays(
                        [
                            pa.compute.equal(self_type_codes, self_i).cast(pa.int8()),
                            pa.array(np.zeros(len(other), dtype=np.int8)),
                        ]
                    ),
                    # offsets=pa.array(np.arange(len(self), dtype=np.int32)),
                    offsets=pa_concat_arrays(
                        [
                            self_offsets,
                            pa.array(np.zeros(len(other), dtype=np.int32)),
                        ]
                    ),
                )
            )
    for other_i in remaining_other_indexes:
        other_member_type = other_field_types[other_i]
        indent_print(depth, "OTHER FIELD solo", other_i, other_member_type)
        other_field = other_fields[other_i]
        new_member_awls.append(
            UnionMember(
                values=ArrowWeaveList(
                    other_field,
                    other_member_type,
                    other._artifact,
                    invalid_reason="Possibly nullable",
                ),
                mask=pa_concat_arrays(
                    [
                        pa.array(np.zeros(len(self), dtype=np.int8)),
                        pa.compute.equal(other_type_codes, other_i).cast(pa.int8()),
                    ]
                ),
                # offsets=pa.array(np.arange(len(other), dtype=np.int32)),
                offsets=pa_concat_arrays(
                    [
                        pa.array(np.zeros(len(self), dtype=np.int32)),
                        other_offsets,
                    ]
                ),
            )
        )

    # I think we will always have at least two union members here. Because we
    # must have had a union on one side coming in.
    if len(new_member_awls) < 2:
        raise errors.WeaveInternalError("Not enough union members after merge")

    member_types = [m.values.object_type for m in new_member_awls]

    result_type_codes = pa.array(np.zeros(len(self) + len(other), dtype=np.int8))
    for i, new_member_awl in enumerate(new_member_awls):
        result_type_codes = pa.compute.add(
            result_type_codes, pa.compute.multiply(new_member_awl.mask, i)
        )
    result_offsets = pa.array(np.zeros(len(self) + len(other), dtype=np.int32))
    for i, new_member_awl in enumerate(new_member_awls):
        result_offsets = pa.compute.add(
            result_offsets,
            pa.compute.multiply(new_member_awl.mask, new_member_awl.offsets),
        )

    assert len(member_types) == len(new_member_awls)
    expected_object_type_members = len(new_member_awls)
    new_object_type = types.UnionType(*member_types)
    if len(new_object_type.members) != expected_object_type_members:
        indent_print(
            depth,
            "ERROR PRODUCING UNION",
            len(new_object_type.members),
            len(new_member_awls),
        )
        indent_print(depth, "new_object.type.members")
        for m in new_object_type.members:
            indent_print(depth + 1, m)
        indent_print(depth, "new_member_awls")
        for m2 in new_member_awls:
            indent_print(depth + 1, m2.values.object_type)

        assert len(new_object_type.members) == expected_object_type_members

    indent_print(depth, "NEW MEMBER TYPE", new_object_type)
    indent_print(depth, "RESULT NULLABLE", result_nullable)
    return ArrowWeaveList(
        pa.UnionArray.from_dense(
            result_type_codes.cast(pa.int8()),
            result_offsets.cast(pa.int32()),
            [m.values._arrow_data for m in new_member_awls],
        ),
        types.optional(new_object_type) if result_nullable else new_object_type,
        self._artifact,
    )


def concatenate(
    self: "ArrowWeaveList", other: "ArrowWeaveList", depth=0
) -> "ArrowWeaveList":
    with unsafe_awl_construction("Possibly nullable"):
        result = _concatenate(self, other, depth)
    result.validate()
    return result


def _merge_concat_split(
    arr: list[ArrowWeaveList],
) -> tuple[list[ArrowWeaveList], list[ArrowWeaveList]]:
    if len(arr) < 2:
        raise ValueError("arr must have length of at least 2")
    middle_index = len(arr) // 2
    return arr[:middle_index], arr[middle_index:]


def concatenate_all(arr: list[ArrowWeaveList]) -> ArrowWeaveList:
    if len(arr) == 0:
        raise ValueError("arr must not be empty")
    if len(arr) == 1:
        return arr[0]
    left, right = _merge_concat_split(arr)
    return concatenate_all(left).concat(concatenate_all(right))



# Official sync/async interface for filesystem operations. All
# weave interactions with the filesystem should go through this
# interface.
#
# Important features:
# - Threadsafe atomic read/write operations
# - Root directory controlled via context for access control
#
# Note: The above is not the case yet! We're in the middle of migrating
# to this interface.

import asyncio
import time
import typing
import contextlib
import os
import aiofiles
import aiofiles.os as aiofiles_os
from aiofiles.threadpool import text as aiofiles_text
from aiofiles.threadpool import binary as aiofiles_binary

from . import engine_trace
from . import errors
from . import util
from . import environment
from . import cache


tracer = engine_trace.tracer()  # type: ignore
async_utime = aiofiles_os.wrap(os.utime)  # type: ignore

# collection for storing strong refs to background tasks to ensure
# they are completely executed (see https://docs.python.org/3/library/asyncio-task.html#asyncio.create_task)
background_tasks: set[asyncio.Task] = set()


def is_subdir(path: str, root: str) -> bool:
    path = os.path.abspath(path)
    root = os.path.abspath(root)
    return os.path.commonpath([path, root]) == root


def safe_path(path: str) -> str:
    root = get_filesystem_dir()
    result = os.path.join(root, path)
    if not is_subdir(result, root):
        raise errors.WeaveAccessDeniedError(f"Path {path} is not allowed")
    return result


class Filesystem:
    def path(self, path: str) -> str:
        return safe_path(path)

    def exists(self, path: str) -> bool:
        return os.path.exists(self.path(path))

    def getsize(self, path: str) -> int:
        return os.path.getsize(self.path(path))

    def makedirs(self, path: str, exist_ok: bool) -> None:
        os.makedirs(self.path(path), exist_ok=exist_ok)

    def touch(self, path: str, newtime: typing.Optional[float] = None) -> None:
        if newtime is None:
            newtime = time.time()
        os.utime(self.path(path), (newtime, newtime))

    def stat(self, path: str) -> os.stat_result:
        return os.stat(self.path(path))

    @contextlib.contextmanager
    def open_write(
        self, path: str, mode: str = "wb"
    ) -> typing.Generator[typing.IO, None, None]:
        path = self.path(path)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        tmp_name = f"{path}.tmp-{util.rand_string_n(16)}"
        with open(tmp_name, mode) as f:
            yield f
        with tracer.trace("rename"):
            os.rename(tmp_name, path)

    @contextlib.contextmanager
    def open_read(
        self, path: str, mode: str = "rb"
    ) -> typing.Generator[typing.IO, None, None]:
        safe_path = self.path(path)
        with open(safe_path, mode) as f:
            if environment.enable_touch_on_read():
                self.touch(path)
            yield f


class FilesystemAsync:
    def path(self, path: str) -> str:
        return safe_path(path)

    async def exists(self, path: str) -> bool:
        return await aiofiles_os.path.exists(self.path(path))

    async def getsize(self, path: str) -> int:
        return await aiofiles_os.path.getsize(self.path(path))

    async def makedirs(self, path: str, exist_ok: bool) -> None:
        await aiofiles_os.makedirs(self.path(path), exist_ok=exist_ok)

    async def touch(self, path: str, newtime: typing.Optional[float] = None) -> None:
        if newtime is None:
            newtime = time.time()
        await async_utime(self.path(path), (newtime, newtime))

    async def stat(self, path: str) -> os.stat_result:
        return await aiofiles_os.stat(self.path(path))

    # Whew! These type shenaningans were tough to figure out!

    @typing.overload
    def open_write(
        self,
        path: str,
        mode: typing.Literal["w"],
    ) -> typing.AsyncContextManager[aiofiles_text.AsyncTextIOWrapper]:
        ...

    @typing.overload
    def open_write(
        self,
        path: str,
        mode: typing.Literal["wb"] = "wb",
    ) -> typing.AsyncContextManager[aiofiles_binary.AsyncBufferedIOBase]:
        ...

    @contextlib.asynccontextmanager
    async def open_write(
        self,
        path: str,
        mode: typing.Union[typing.Literal["w"], typing.Literal["wb"]] = "wb",
    ) -> typing.Any:
        path = self.path(path)
        await aiofiles_os.makedirs(os.path.dirname(path), exist_ok=True)
        tmp_name = f"{path}.tmp-{util.rand_string_n(16)}"
        async with aiofiles.open(tmp_name, mode) as f:
            yield f
        with tracer.trace("rename"):
            await aiofiles_os.rename(tmp_name, path)

    @typing.overload
    def open_read(
        self,
        path: str,
        mode: typing.Literal["r"],
    ) -> typing.AsyncContextManager[aiofiles_text.AsyncTextIOWrapper]:
        ...

    @typing.overload
    def open_read(
        self,
        path: str,
        mode: typing.Literal["rb"] = "rb",
    ) -> typing.AsyncContextManager[aiofiles_binary.AsyncBufferedIOBase]:
        ...

    @contextlib.asynccontextmanager
    async def open_read(
        self,
        path: str,
        mode: typing.Union[typing.Literal["r"], typing.Literal["rb"]] = "rb",
    ) -> typing.Any:
        safe_path = self.path(path)
        async with aiofiles.open(safe_path, mode) as f:
            if environment.enable_touch_on_read():
                now = time.time()  # ensure the new atime is from before the yield
                # fire and forget, dont block yield of f
                task = asyncio.create_task(self.touch(path, newtime=now))
                background_tasks.add(task)
                task.add_done_callback(background_tasks.discard)
            yield f


def get_filesystem() -> Filesystem:
    return Filesystem()


def get_filesystem_async() -> FilesystemAsync:
    return FilesystemAsync()


def get_filesystem_dir() -> str:
    path = environment.weave_filesystem_dir()
    prefix = cache.get_cache_prefix_context()
    if prefix is not None:
        path = os.path.join(path, prefix)
    cache_key = cache.get_user_cache_key()
    if cache_key is not None:
        return os.path.join(path, cache_key)
    return path



import json
import typing
from dataclasses import dataclass, field

from . import weave_types as types
from . import artifact_fs


T = typing.TypeVar("T", bound="PartialObject")


class PartialObjectTypeGeneratorType(types._PlainStringNamedType):
    """Base class for types like projectType(), runType(), etc. Instances of this class do not have keys,
    but they have a method called with_keys() that allows them to generate instances of the type with
    keys. E.g.,

    projectType().with_keys({"name": String()}) -> projectTypeWithKeys({"name": String()})

    Assignability rules:
    --------------------

    projectType() <- projectTypeWithKeys({"name": String()})  VALID
    projectTypeWithKeys({"name": Int()}) <- projectType()     INVALID
    """

    @classmethod
    def with_keys(cls, attrs: dict[str, types.Type]) -> "PartialObjectType":
        """Creates a new Weave Type that is assignable to the original Weave Type, but
        also has the specified keys. This is used during the compile pass for creating a Weave Type
        to represent the exact shape of a PartialObject, and for communicating that shape to arrow.
        """

        return PartialObjectType(cls, attrs)

    @classmethod
    def type_of_instance(cls, obj: "PartialObject") -> types.Type:
        keys_type = typing.cast(types.TypedDict, types.TypeRegistry.type_of(obj.keys))
        return cls.with_keys(keys_type.property_types)


@dataclass
class PartialObject:
    keys: dict = field(default_factory=dict)

    @classmethod
    def from_keys(cls: typing.Type[T], key_dict: dict) -> T:
        return cls(key_dict)

    def __getitem__(self, key: str) -> typing.Any:
        return self.keys[key]

    def get(self, key: str, default: typing.Any = None) -> typing.Any:
        return self.keys.get(key, default)


class PartialObjectType(types.Type):
    """Base class for types like projectTypeWithKeys({...}), runTypeWithKeys({...}), etc.

    Assignability rules:
    --------------------

    projectType() <- projectTypeWithKeys({"name": String()})  VALID
    projectTypeWithKeys({"name": Int()}) <- projectType()     INVALID
    """

    # e.g., Project.WeaveType. Note: this property is the class itself, not an instance of the class.
    keyless_weave_type_class: typing.Type[types.Type]
    keys: dict[str, types.Type]

    def __init__(
        self,
        keyless_weave_type_class: typing.Type[types.Type],
        keys: typing.Union[dict[str, types.Type], types.TypedDict],
    ):
        self.keyless_weave_type_class = keyless_weave_type_class
        if isinstance(keys, types.TypedDict):
            keys = keys.property_types
        self.keys = keys

    def _assign_type_inner(self, other_type: types.Type) -> bool:
        return (
            isinstance(other_type, PartialObjectType)
            and self.keyless_weave_type_class == other_type.keyless_weave_type_class
            and self.keys == other_type.keys
        )

    def _is_assignable_to(self, other_type: types.Type) -> typing.Optional[bool]:
        if other_type.__class__ is self.keyless_weave_type_class:
            return True

        if isinstance(other_type, PartialObjectType):
            if self.keyless_weave_type_class != other_type.keyless_weave_type_class:
                return False

            other_td = types.TypedDict(other_type.keys)
            self_td = types.TypedDict(self.keys)
            return other_td.assign_type(self_td)

        return False

    def _str_repr(self) -> str:
        keys_repr = dict(sorted([(k, v.__repr__()) for k, v in self.keys.items()]))
        return f"{self.keyless_weave_type_class.__name__}({keys_repr})"

    def __repr__(self) -> str:
        return self._str_repr()

    def _to_dict(self) -> dict:
        property_types = {}
        for key, type_ in self.keys.items():
            property_types[key] = type_.to_dict()
        result = {
            "keys": property_types,
            "keyless_weave_type_class": self.keyless_weave_type_class().to_dict(),
        }
        return result

    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, PartialObjectType):
            return (
                self.keyless_weave_type_class == other.keyless_weave_type_class
                and self.keys == other.keys
            )
        return False

    @classmethod
    def from_dict(cls, d: dict) -> "PartialObjectType":
        property_types = {}
        for key, type_ in d["keys"].items():
            property_types[key] = types.TypeRegistry.type_from_dict(type_)
        keyless_weave_type_class = types.TypeRegistry.type_from_dict(
            d["keyless_weave_type_class"]
        ).__class__

        return cls(keyless_weave_type_class, property_types)

    def save_instance(
        self, obj: PartialObject, artifact: artifact_fs.FilesystemArtifact, name: str
    ) -> None:
        from . import mappers_python

        serializer = mappers_python.map_to_python(self, artifact)
        result = serializer.apply(obj)
        with artifact.new_file(
            f"{name}.{self.keyless_weave_type_class.__name__}WithKeys.json"
        ) as f:
            json.dump(result, f, allow_nan=False)

    def load_instance(
        self,
        artifact: artifact_fs.FilesystemArtifact,
        name: str,
        extra: typing.Optional[list] = None,
    ) -> PartialObject:
        from . import mappers_python

        with artifact.open(
            f"{name}.{self.keyless_weave_type_class.__name__}WithKeys.json"
        ) as f:
            result = json.load(f)
        mapper = mappers_python.map_from_python(self, artifact)
        mapped_result = mapper.apply(result)
        if extra is not None:
            return mapped_result[extra[0]]
        return typing.cast(PartialObject, mapped_result)

    def instance_to_dict(self, obj: PartialObject) -> dict[str, typing.Any]:
        return obj.keys

    def instance_from_dict(self, d: dict[str, typing.Any]) -> PartialObject:
        instance_class = typing.cast(
            typing.Type[PartialObject], self.keyless_weave_type_class.instance_class
        )
        return instance_class.from_keys(d)



import typing
from typing import Iterable, Sequence, Optional
import weakref
import hashlib
import json
import functools

from . import uris
from . import box
from . import errors
from . import graph_client_context
from . import weave_types as types
from . import object_context
from . import context_state
from .language_features.tagging import tag_store

# We store Refs here if we can't attach them directly to the object
REFS: weakref.WeakValueDictionary[int, "Ref"] = weakref.WeakValueDictionary()

if typing.TYPE_CHECKING:
    from . import weave_types as types
    from . import run


def _map_to_ref_strs(obj: typing.Any) -> typing.Any:
    if isinstance(obj, dict):
        return {k: _map_to_ref_strs(v) for k, v in obj.items()}  # type: ignore
    if isinstance(obj, list):
        return [_map_to_ref_strs(v) for v in obj]  # type: ignore
    ref = get_ref(obj)
    if ref is not None:
        return str(ref)
    return obj


class Ref:
    _obj: typing.Any
    _type: typing.Optional["types.Type"]
    extra: typing.Optional[list[str]]

    def __init__(
        self,
        obj: typing.Optional[typing.Any] = None,
        type: typing.Optional["types.Type"] = None,
        extra: typing.Optional[list[str]] = None,
    ):
        self._type = type
        self.extra = extra
        if obj is not None and type is not None and type.name != "tagged":  # type: ignore
            obj = box.box(obj)
            _put_ref(obj, self)
        self._obj = obj

    @property
    def obj(self) -> typing.Any:
        obj_ctx = object_context.get_object_context()
        if obj_ctx:
            val = obj_ctx.lookup_ref_val(str(self.initial_uri))
            if not isinstance(val, object_context.NoResult):
                return val

        if self._obj is not None:
            return self._obj

        if not self.is_saved:
            # PR TODO: this path needs to happen in FSArtifact, as you can
            # always get the value of a MemArtifact
            # PR: I changed this back to None to get tests passing. What breaks
            # now? My guess is mutations from the UI? Or maybe creating brand
            # new objects from the UI (datasets)
            return None
            raise errors.WeaveArtifactVersionNotFound

        if self.extra is None:
            obj = self._get()
            obj = box.box(obj)
        else:
            ref_without_extra = self.without_extra(None)
            outer_obj = ref_without_extra.get()
            try:
                with context_state.ref_tracking(True):
                    obj = outer_obj._lookup_path(self.extra)  # type: ignore
            except AttributeError:
                raise errors.WeaveInternalError(
                    f"Ref has extra {self.extra} but object of type {type(outer_obj)} does not support _lookup_path"
                )

        # Don't put a ref if the object is tagged, it leads to failures in a couple
        # of the test_arrow.py tests. I don't have the exact rationale, but it's something
        # like we end up later loading the tagged object when we want the untagged one,
        # of vice versa.
        # this check was if self.type.name != 'tagged', however
        # accesing self.type here is problematic, can cause a loop (as it does with tableref).
        # Since we just loaded this object, it will be tagged if it was stored tagged,
        # which I believe is the check we want.
        if not tag_store.is_tagged(obj):
            _put_ref(obj, self)
        self._obj = obj

        if obj_ctx is not None:
            obj_ctx.add_ref(str(self.initial_uri), obj, self.type)

        return obj

    @property
    def ui_url(self) -> str:
        return "[no url for obj]"

    @property
    def type(self) -> "types.Type":
        raise NotImplementedError

    @functools.cached_property
    def digest(self) -> str:
        hash = hashlib.md5()
        # This can encounter non-serialized objects, even though Ref
        # must be a pointer to an object that can only contain serializable
        # stuff and refs. But we recursively deserialize all sub-refs when
        # fetching a ref. So we need to walk obj, converting objs back to
        # refs where we can, before this json.dumps call.
        # TODO: fix
        with_refs = _map_to_ref_strs(self.obj)
        hash.update(json.dumps(with_refs).encode())
        return hash.hexdigest()

    @classmethod
    def from_str(cls, s: str) -> "Ref":
        return uris.WeaveURI.parse(s).to_ref()

    @property
    def initial_uri(self) -> str:
        raise NotImplementedError

    @property
    def uri(self) -> str:
        raise NotImplementedError

    @property
    def is_saved(self) -> bool:
        raise NotImplementedError

    def get(self) -> typing.Any:
        return self.obj

    def _get(self) -> typing.Any:
        raise NotImplementedError

    def without_extra(self, new_type: typing.Optional[types.Type]) -> "Ref":
        raise NotImplementedError

    def with_extra(
        self, new_type: typing.Optional[types.Type], obj: typing.Any, extra: list[str]
    ) -> "Ref":
        raise NotImplementedError

    def __str__(self) -> str:
        return str(self.uri)

    def input_to(self) -> Sequence["run.Run"]:
        client = graph_client_context.require_graph_client()
        return client.ref_input_to(self)

    def value_input_to(self) -> Sequence["run.Run"]:
        client = graph_client_context.require_graph_client()
        return client.ref_value_input_to(self)

    def output_of(self) -> Optional["run.Run"]:
        client = graph_client_context.require_graph_client()
        return client.ref_output_of(self)


def get_ref(obj: typing.Any) -> typing.Optional[Ref]:
    if isinstance(obj, Ref):
        return obj
    if hasattr(obj, "_ref"):
        return obj._ref
    try:
        if id(obj) in REFS:
            return REFS[id(obj)]
    except TypeError:
        pass
    return None


# TODO: this cannot be used on tagged objects, since we don't know
# whether obj is intended to be the tagged or untagged version
# So for now we never call it on tagged objects. But this probably
# breaks some behaviors (like automatic cross-artifact references
# for tagged objects)
def _put_ref(obj: typing.Any, ref: Ref) -> None:
    try:
        obj._ref = ref
    except (AttributeError, ValueError):
        if isinstance(obj, (int, float, str, list, dict, set)):
            return
        REFS[id(obj)] = ref


def clear_ref(obj: typing.Any) -> None:
    try:
        delattr(obj, "_ref")
    except AttributeError:
        pass
    if id(obj) in REFS:
        REFS.pop(id(obj))


def deref(ref: typing.Any) -> typing.Any:
    if isinstance(ref, Ref):
        return ref.get()
    return ref


types.RefType.instance_classes = Ref



import typing
import pyarrow as pa
from pyarrow import compute as pc
from .. import weave_types as types
from .. import graph


# Reimplementation of Weave0 `toSafeCall` which
# converts media to their digest
def _to_compare_safe_call(node: graph.OutputNode) -> graph.OutputNode:
    from ..ops_primitives.dict import dict_
    from ..ops_domain.wbmedia import ArtifactAssetType

    node_type = types.non_none(node.type)
    if ArtifactAssetType.assign_type(node_type):
        # Must add this to all assets before merging
        return node.sha256  # type: ignore
    elif types.TypedDict({}).assign_type(node_type):
        new_keys = {}
        dirty = False
        for key in node_type.property_types.keys():  # type: ignore
            sub_key = node[key]  # type: ignore
            new_val = _to_compare_safe_call(sub_key)
            new_keys[key] = new_val
            if new_val is not sub_key:
                dirty = True
        if dirty:
            return dict_(**new_keys)
    elif types.List().assign_type(node_type):
        # AWL does not like joining on lists.
        return node.joinToStr(",")  # type: ignore
    return node


def _eq_null_consumer_helper(
    lhs: pa.Array, rhs: typing.Union[pa.Array, pa.Scalar]
) -> typing.Tuple[pa.Array, pa.Array]:
    # consume nulls

    self_is_null = pc.is_null(lhs)
    other_is_null = pc.is_null(rhs)
    one_null = pc.xor(self_is_null, other_is_null)
    both_null = pc.and_(self_is_null, other_is_null)

    return one_null, both_null


def not_equal(lhs: pa.Array, rhs: typing.Union[pa.Array, pa.Scalar]) -> pa.Array:
    # this unboxes the null scalar if it is boxed
    if rhs == None:
        rhs = None

    if isinstance(lhs, pa.NullArray) and (isinstance(rhs, pa.NullArray) or not rhs):
        return pc.fill_null(pc.cast(lhs, pa.bool_()), False)

    one_null, both_null = _eq_null_consumer_helper(lhs, rhs)
    result = pc.not_equal(lhs, rhs)
    result = pc.replace_with_mask(result, both_null, False)
    result = pc.replace_with_mask(result, one_null, True)
    return result


def equal(lhs: pa.Array, rhs: typing.Union[pa.Array, pa.Scalar]) -> pa.Array:
    # this unboxes the null scalar if it is boxed
    if rhs == None:
        rhs = None

    one_null, both_null = _eq_null_consumer_helper(lhs, rhs)
    result = pc.equal(lhs, rhs)
    result = pc.replace_with_mask(result, both_null, True)
    result = pc.replace_with_mask(result, one_null, False)
    return result



from ..op_def import OpDef
from ..op_def_type import OpDefType
from .. import weave_types as types
from .. import api as weave


@weave.weave_class(weave_type=OpDefType)
class OpDefNodeMethods:
    # Unfortunate: we can't call this "name" because name is a
    # VarNode attribute. TODO: Fix
    @weave.op()
    def op_name(self) -> str:
        return self.simple_name  # type: ignore

    @weave.op()
    def output_type(self) -> types.Type:
        if callable(self.output_type):
            return types.Invalid()
        return self.output_type



import hashlib
import typing
from typing import Mapping
import json
import dataclasses
import random

from . import storage
from . import ref_base
from . import op_def
from . import weave_types as types
from . import runs
from . import graph
from . import artifact_local
from . import weave_internal
from . import op_policy


@dataclasses.dataclass
class RunKey:
    op_simple_name: str
    id: str


def _value_id(val):
    # Important, do not include the type here, as it can change.
    # This happens because you can have a ref to an item that's in a list.
    # The list's object_type can change as items are appended to it.
    # We don't know the specific type of each item within the list without
    # further refinement.
    hash_val = json.dumps(storage.to_python(val)["_val"])
    hash = hashlib.md5()
    hash.update(json.dumps(hash_val).encode())
    return hash.hexdigest()


def make_run_key(
    op_def: op_def.OpDef,
    inputs_refs: Mapping[str, typing.Any],
    impure_cache_key: typing.Optional[str] = None,
) -> RunKey:
    hash_val: typing.Any
    if not op_def.pure and impure_cache_key is None:
        hash_val = random.random()
    else:
        hashable_inputs = {}
        for name, obj in inputs_refs.items():
            hashable_inputs[name] = _value_id(obj)
        hash_val = {
            "op_name": op_def.name,
            "op_version": op_def.version,
            "inputs": hashable_inputs,
        }
        if impure_cache_key is not None:
            hash_val["impure_cache_key"] = impure_cache_key
    hash = hashlib.md5()
    hash.update(json.dumps(hash_val).encode())

    # For now, put op_def name in the run ID. This makes debugging much
    # easier because you can inspect the local artifact directly names.
    # This may not be what we want in production.
    return RunKey(op_def.simple_name, hash.hexdigest())


# Trace interface. Makes use of objects and mutations to store trace data.
# Manually constructs nodes and op calls to avoid recursively calling
# the execute engine, either via use or type refinement.
class TraceLocal:
    def new_run(
        self,
        run_key: RunKey,
        inputs: typing.Optional[dict[str, ref_base.Ref]] = None,
        output: typing.Any = None,
    ) -> graph.Node[runs.Run]:
        run = runs.Run(run_key.id, run_key.op_simple_name)
        if inputs is not None:
            run.inputs = inputs
        if output is not None:
            run.output = output
        self.save_run(run)
        return self.get_run(run_key)

    def _single_run(self, run_key: RunKey) -> graph.Node[runs.Run]:
        single_uri = artifact_local.WeaveLocalArtifactURI(
            f"run-{run_key.op_simple_name}-{run_key.id}", "latest", "obj"
        )
        return weave_internal.manual_call(
            "get",
            {"uri": graph.ConstNode(types.String(), str(single_uri))},
            types.RunType(),
        )

    def _run_table(self, run_key: RunKey):
        table_uri = artifact_local.WeaveLocalArtifactURI(
            f"run-{run_key.op_simple_name}", "latest", "obj"
        )
        return weave_internal.manual_call(
            "get",
            {"uri": graph.ConstNode(types.String(), str(table_uri))},
            types.List(types.RunType()),
        )

    def _should_save_to_table(self, run_key: RunKey):
        # Restricted to just a couple ops for now.
        # A NOTE: for the future, async ops definitely can't be saved to
        # table (until we have safe table writes)
        return op_policy.should_table_cache(run_key.op_simple_name)

    def get_run(self, run_key: RunKey) -> graph.Node[runs.Run]:
        if self._should_save_to_table(run_key):
            return weave_internal.manual_call(
                "listobject-lookup",
                {
                    "arr": self._run_table(run_key),
                    "id": graph.ConstNode(types.String(), run_key.id),
                },
                types.RunType(),
            )
        return self._single_run(run_key)

    def get_run_val(self, run_key: RunKey) -> typing.Optional[runs.Run]:
        from . import execute_fast

        res = execute_fast._execute_fn_no_engine(None, None, self.get_run(run_key))
        return res

    def save_run(self, run: runs.Run):
        from .ops_primitives import weave_api

        run_key = RunKey(run.op_name, run.id)
        if self._should_save_to_table(run_key):
            weave_api.append(self._run_table(run_key), run, {})
        else:
            weave_api.set(self._single_run(run_key), run, {})

    def save_run_output(self, od: op_def.OpDef, run_key: RunKey, output: typing.Any):
        if not od.pure:
            # If an op is impure, its output is saved to a name that does not
            # include run ID. This means consuming pure runs will hit cache if
            # the output of an impure op is the same as it was last time.
            # However that also means we can traceback through impure ops if we want
            # to see the actual query that run for a given object.
            # TODO: revisit this behavior
            return self.save_object(output)
        # TODO: table caching is currently disabled, but this path doesn't handle it
        # when we turn it back on!
        return self.save_object(
            output, name=f"run-{run_key.op_simple_name}-{run_key.id}-output"
        )

    def save_object(
        self, obj: typing.Any, name: typing.Optional[str] = None
    ) -> ref_base.Ref:
        return storage.save(obj, name=name, branch="latest")
